"""
    Selective.PropOptimizer.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from time import sleep
from importlib import import_module, reload
from molass_legacy.Models.ElutionCurveModels import EGH, EMG, EGHA, EMGA
from molass_legacy.Models.RateTheory.EDM import EDM
from molass_legacy.Models.Stochastic.Tripore import Tripore
from .PropOptimizerImpl import AFFINE

IMPL_DICT = {
    'EGH' : 'Selective.PropOptimizerEGH',
    'EMG' : 'Selective.PropOptimizerEMG',
    'EDM' : 'Selective.PropOptimizerEDM',
    'STC' : 'Selective.PropOptimizerSTC',
    }

if AFFINE:
    # this is consistent with traditional V1 models
    MODEL_DICT = {
        'EGH' : EGHA(),
        'EMG' : EMGA(),
        'EDM' : EDM(delayed=True),
        'STC' : Tripore(delayed=True),
        }
else:
    MODEL_DICT = {
        'EGH' : EGH(),
        'EMG' : EMG(),
        'EDM' : EDM(),
        'STC' : Tripore(),
        }

def load_func(modelname, funcname, devel=True):
    classname = IMPL_DICT.get(modelname)
    assert classname is not None

    module = import_module(classname)
    if devel:
        reload(module)
    func = getattr(module, funcname)    # from Selective.PropOptimizerXXX import funcname
    return func

class PropOptimizer:
    def __init__(self, modelname, x, y):
        self.guess_init_params_impl = load_func(modelname, 'guess_init_params')
        self.compute_props_impl = load_func(modelname, 'compute_props')
        self.compute_cy_list_impl = load_func(modelname, 'compute_cy_list')
        self.optimize_to_props_impl = load_func(modelname, 'optimize_to_props')
        self.model = MODEL_DICT.get(modelname)
        self.x = x
        self.y = y
        self.init_params = self.guess_init_params_impl(x, y)

    def get_init_params(self):
        return self.init_params

    def compute_props(self, params):
        return self.compute_props_impl(self.x, params)

    def compute_cy_list(self, params):
        return self.compute_cy_list_impl(self.x, params)

    def optimize(self, *o_args, init_params=None):
        if init_params is None:
            init_params = self.init_params
        return self.optimize_to_props_impl(self.x, self.y, init_params, *o_args)

    def make_xr_opt_recs(self, fx, y, peaks):
        return self.model.make_xr_opt_recs(fx, y, peaks)

    def make_uv_opt_recs(self, fx, uv_y, peaks, scale):
        return self.model.make_uv_opt_recs(fx, uv_y, peaks, scale)

def compute_optimal_proportion_impl(progress_queue, job_args):
    func = load_func(job_args.modelname, 'compute_optimal_proportion')
    func(progress_queue, job_args)

def compute_optimal_proportion(progress_queue, job_args, debug=True):
    from molass_legacy.KekLib.DebugPlot import exec_in_threaded_mainloop, quit_threaded_mainloop

    if debug:
        def exec_closure():
            compute_optimal_proportion_impl(progress_queue, job_args)
            quit_threaded_mainloop()

        exec_in_threaded_mainloop(exec_closure)
    else:
        compute_optimal_proportion_impl(progress_queue, job_args)



"""
    Models.Stochastic.PaperFelinger2004.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
sns.set_theme()

def felinger_sdm_cf(w, n1, t1, Nd, t0):
    return np.exp(Nd * (1 - np.sqrt(1 - 2*1j*w/Nd*(n1*t1/(1 - 1j*t1*w) + t0))))

def felinger_sdm_cf_proof():
    from molass_legacy.Models.ElutionModelUtils import compute_4moments
    from molass_legacy.Models.RateTheory.EDM import guess_params_from_sdm
    from SecTheory.Edm import guess_single_edm, edm_func
    from SecTheory.SecCF import dispersive_monopore
    from SecTheory.SecPDF import FftInvPdf
    dispersive_monopore_pdf = FftInvPdf(dispersive_monopore)
    felinger_sdm_pdf_ = FftInvPdf(felinger_sdm_cf)

    x = np.arange(300)
    t0 = 50
    Nd = 14400
    n1 = 400
    t1 = 0.3

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
    fig.suptitle("Numerical Proof of Equivalence among Dondi, Felinger SDM and Rehman EDM")
    ax1.set_title("Felinger-1999 and Dondi-2002 SDM Formulae")
    y_dondi = dispersive_monopore_pdf(x, n1, t1, Nd, t0)
    ax1.plot(x, y_dondi, label="Dondi-2002 SDM")
    ax1.plot(x, felinger_sdm_pdf_(x, n1, t1, Nd, t0), ":", label="Felinger-1999 SDM")
    ax1.legend()
    params = guess_params_from_sdm(x, y_dondi, n1, t1, Nd, t0)
    print("params=", params)
    # t0, u, a, b, e, Dz, cinj
    t0, u, a, b, e, Dz, cinj = params
    edm_y = edm_func(x-t0, u, a, b, e, Dz, cinj)

    ax2.set_title("Dondi-2002 SDM and Rehman-2021 EDM Formulae")
    ax2.plot(x, y_dondi, label="Dondi-2002 SDM")
    ax2.plot(x, edm_y, ":", label="Rehman-2021 EDM")

    ax2.legend()
    fig.tight_layout()
    plt.show()

def demo():
    from molass_legacy.Models.ElutionCurveModels import egh
    from SecTheory.Edm import guess_single_edm, edm_func

    # t, u, a, b, e, Dz, cinj
    x = np.arange(300)
    egh_y = egh(x, 1, 120, 20, 20)
    edm = guess_single_edm(x, egh_y)
    params = edm.get_comp_params()
    print("params=", params)
    u = 0.5
    edm_y = edm_func(x, u, *params)

    fig, ax = plt.subplots()
    ax.plot(x, egh_y, label="EGH")
    ax.plot(x, edm_y, ":",  label="EDM")
    ax.legend()
    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    sys.path.append("../lib")

    felinger_sdm_cf_proof()
    # demo()


"""
This script edits all .rst files in the docs directory to update the title lines.

See also: `"Modify *.rst files" <LINK1_>`__.

.. _LINK1: https://nshimizu0721.github.io/molass-develop/chapters/07/documentation.html#modify-rst-files

"""
import os

docs_dir = "./source"

for root, _, files in os.walk(docs_dir):
    for file in files:
        if file.endswith(".rst"):
            file_path = os.path.join(root, file)
            with open(file_path, "r", encoding="utf-8") as f:
                content = f.readlines()

            # Update the title line
            updated_content = []
            for line in content:
                if line.startswith(r"molass\_legacy."):
                    line = line.replace(r"molass\_legacy.", "").replace(" package", "").replace(" module", "")
                elif line.startswith("Subpackages"):
                    line = line.replace("Subpackages", "Submodules")
                updated_content.append(line)

            # Write the updated content back to the file
            with open(file_path, "w", encoding="utf-8") as f:
                f.writelines(updated_content)

print("All .rst files have been updated.")
# This script will walk through all the .rst files in the docs directory and update the title lines as specified.


# coding: utf-8
"""

    ScatteringBasespline.py

    scattering baseline solver

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import numpy            as np
from scipy                  import stats
from scipy.interpolate      import UnivariateSpline
from bisect                 import bisect_right
from molass_legacy.Baseline.ScatteringBaseline     import ScatteringBaseline

DEBUG_PLOT              = False
PERCENTILE_SECOND       = 25
PERCENTILE_FINAL        = 10

VERY_SMALL_ANGLE_LIMIT  = 1.0
NUM_REGRESSION_POINTS   = 20

def get_spline_smoothing_for_base( y, max_y ):
    x   = np.arange(NUM_REGRESSION_POINTS)
    slope, intercept, r_value, p_value, std_err = stats.linregress( x, y[0:NUM_REGRESSION_POINTS] )
    s   = std_err**2 * 1e3 * len( y )
    # print( 'std_err=', std_err )
    # print( 'max_y=', max_y, 's=%.1e' % s )
    return s

class ScatteringBasespline:
    def __init__( self, y, q = 0.02, q_min=0.004 ):
        self.q  = q
        self.q_min = q_min
        self.y  = y
        self.x = np.arange( len(self.y) )
        self.line   = ScatteringBaseline( y )
        # np.savetxt( 'curve.csv', y, delimiter=',' )

    def solve( self, p_final=PERCENTILE_FINAL ):
        A, B = self.line.solve( p_final=p_final )
        # A, B = self.line.solve( p_final=PERCENTILE_FINAL )
        self.A  = A
        self.B  = B
        self.baseline = self.A*self.x + self.B

        if self.q < VERY_SMALL_ANGLE_LIMIT:
            max_y   = np.max( self.y )
            smoothing   = get_spline_smoothing_for_base( self.y, max_y )
            y_  = self.y - self.baseline
            ypps   = np.percentile( y_, [ PERCENTILE_SECOND, p_final ] )
            self.lowp0 = y_ < ypps[0]
            self.curve  = UnivariateSpline( self.x[self.lowp0], self.y[self.lowp0], s=smoothing )
            # TODO: shift using p_final (the following is not tested and probably not correct )
            self.final_shift = ypps[0] - ypps[1]
            print( 'self.final_shift=', self.final_shift )

        return 0, 0, 0

    def get_baseline( self, x ):
        baseline    = self.baseline
        if self.q < VERY_SMALL_ANGLE_LIMIT:
            # basecurve = self.curve( x ) + self.curve_base
            basecurve = self.curve( x ) + self.final_shift
            # w = VERY_SMALL_ANGLE_LIMIT - self.q_min
            # composite = basecurve * ( VERY_SMALL_ANGLE_LIMIT - self.q )/w + baseline * ( self.q - self.q_min )/w
            if DEBUG_PLOT:
                from DebugCanvas    import DebugCanvas
                def debug_plot( fig ):
                    ax = fig.add_subplot( 111 )
                    ax.set_title( 'q=' + str(self.q) )
                    ax.plot( self.y, label='data' )
                    ax.plot( baseline, label='baseline' )
                    ax.plot( basecurve, label='basecurve' )
                    ax.plot( self.x[self.lowp0], self.y[self.lowp0], 'yo' )
                    knots = self.curve.get_knots()
                    ax.plot( knots, self.curve(knots), 'co' )
                    # ax.plot( composite, label='composite' )
                    # ax.plot( self.lowp1, self.y[self.lowp1], 'o', color='red' )
                    ax.legend()
                    fig.tight_layout()
                dc = DebugCanvas( "Debug", debug_plot, parent=None, figsize=(8,6) )
                dc.show( cancelable=True )
            return basecurve
        else:
            return baseline



"""
    square_root_transformation.py

    Copyright (c) 2016, Masatsuyo Takahashi
"""
import numpy as np
import pylab as plt

delta = 1e-2

intervals = map( lambda xi: np.arange( xi[0], xi[-1], delta ), [ [ 0.5, 0.6 ], [ 0.95, 1.05 ], [ 1.4, 1.5 ] ] )

x = np.arange( 0, 3, delta )

y = 2*x - 1

fig = plt.figure( figsize=(8,8) )
ax = fig.add_subplot( 1, 1, 1 )
ax.set_title( 'Comparison:  $2x-1$ vs $\sqrt{2x-1}$' )

ax.set_xlim( [ -0.2, 2.2 ] )
ax.set_ylim( [ -0.2, 2.2 ] )

xmin, xmax = ax.get_xlim()
ymin, ymax = ax.get_ylim()

xoffset = ( xmax - xmin ) * 0.05
yoffset = ( ymax - ymin ) * 0.05

ax.plot( [ xmin, xmax ], [ 0, 0 ], color='black' )
ax.plot( [ 0, 0 ], [ ymin, ymax ], color='black' )

ax.plot( x, y, color='blue' )

y_ = np.sqrt( 2*x - 1 )

ax.plot( x, y_, color='green' )

for xi in intervals:
    yL = 2*xi - 1
    yS = np.sqrt( yL )
    yZ = np.zeros( ( len(xi) ) )
    xZ = np.zeros( ( len(xi) ) )
    for i, x_ in enumerate( [ xi[0], xi[-1] ] ):
        ax.plot( [ x_, x_ ], [ ymin, ymax ], ':', color='red' )
        ax.plot( [ 0, x_ ], [ yL[-i], yL[-i] ], ':', color='blue' )
        ax.plot( [ 0, x_ ], [ yS[-i], yS[-i] ], ':', color='green' )

    ax.plot( xi, yZ, color='red', linewidth=7, alpha=0.5 )
    ax.plot( xi, yL, color='blue', linewidth=7, alpha=0.5 )
    ax.plot( xZ, yL, color='blue', linewidth=7, alpha=0.5 )
    ax.plot( xi, yS, color='green', linewidth=7, alpha=0.5 )
    ax.plot( xZ, yS, color='green', linewidth=7, alpha=0.5 )

for x_ in [ 0.6, 1, 1.5 ]:
    yy = np.sqrt( 2*x_ - 1 )
    yd = 1/np.sqrt( 2*x_ - 1 )
    text_ = 'slope is %.2g' % yd
    ax.annotate( text_, xy=( x_, yy ),
        xytext=( x_ + xoffset, yy - yoffset ), alpha=0.5,
        arrowprops=dict( headwidth=5, width=1, facecolor='black', shrink=0, alpha=0.5 ),
        )

plt.show()



"""
    Theory.Conjugate.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt

def compute_conjugate_solutin(M, C, aq, bq):
    def obj_func(p):
        P_ = np.array([p, -bq]).T
        return np.linalg.norm(P_@C - M)

    res = minimize(obj_func, aq)
    return res.x, res.fun

def spike(preview_results):
    from SimpleGuinier import SimpleGuinier

    A1 = np.loadtxt(preview_results + '/A1.dat')
    B1 = np.loadtxt(preview_results + '/B1.dat')
    C1 = np.loadtxt(preview_results + '/C1.dat')
    M1 = np.loadtxt(preview_results + '/M1.dat')
    a1 = A1[:,1]
    b1 = B1[:,1]
    P1 = np.array([a1, b1]).T
    c = C1[0,:]
    C = np.array([c, c**2])
    norm1 = np.linalg.norm(P1@C - M1)
    print(norm1)

    a1_, minnorm = compute_conjugate_solutin(M1, C, a1, b1)
    print(minnorm, abs(minnorm - norm1)/np.linalg.norm(M1))

    qv = A1[:,0]
    sg = SimpleGuinier(np.array([qv, a1_, A1[:,2]]).T)
    print('Rg=', sg.Rg)

    fig, (ax1, ax2)= plt.subplots(ncols=2, figsize=(14, 7))

    ax1.set_yscale('log')
    ax1.plot(qv, a1, label='$A(q)$')
    ax1.plot(qv, a1_, label='$A^-(q)$')

    ax2.plot(qv, b1, label='$B(q)$')
    ax2.plot(qv, -b1, label='$-B(q)$')

    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    plt.show()



# coding: utf-8
"""
    MdViewer.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from MatrixData import compute_3d_xyz

class MdViewer(Dialog):
    def __init__(self, parent, D, xvector=None):
        self.D = D
        if xvector is None:
            xvector = np.arange(D.shape[0])
        self.xvector = xvector
        self.yvector = np.arange(D.shape[1])
        self.xones = np.ones(D.shape[1])
        self.i = bisect_right(xvector, 0.02)
        self.artists = None
        Dialog.__init__(self, parent, "MdViewer", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        self.fig = fig = plt.figure(figsize=(14,7))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        gs = GridSpec(1,2)
        ax1 = fig.add_subplot(gs[0,0], projection='3d')
        ax2 = fig.add_subplot(gs[0,1])
        axs = fig.add_axes([0.05, 0.01, 0.4, 0.03])
        self.axes = [ax1, ax2, axs]

        xx, yy, zz = compute_3d_xyz(self.D, x=self.xvector)
        ax1.plot_surface(xx, yy, zz, alpha=0.5)

        self.draw_elution()
        self.slider = Slider(axs, 'X', self.xvector[0], self.xvector[-1], valinit=self.xvector[self.i])

        self.slider.on_changed(self.on_x_change)

        fig.tight_layout()
        self.mpl_canvas.draw()

    def on_x_change(self, val):
        # print(val)
        self.i = bisect_right(self.xvector, val)
        self.draw_elution()
        self.mpl_canvas.draw()

    def draw_elution(self):
        ax1, ax2 = self.axes[0:2]
        q = self.xvector[self.i]
        x = self.xones*q
        y = self.yvector
        z = self.D[self.i,:]
        if self.artists is None:
            eline3d, = ax1.plot(x, y, z, color='orange')
            eline2d, = ax2.plot(y, z, color='orange')
            self.artists = [eline3d, eline2d]
        else:
            eline3d, eline2d = self.artists
            eline3d.set_data(x, y)
            eline3d.set_3d_properties(z)
            eline2d.set_data(y, z)

        zmin = np.min(z)
        zmax = np.max(z)
        ax2.set_ylim(zmin, zmax)



"""
    Alsaker.ComparisonPlot.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def plot_comparison_result_impl(corrected_sd, comparison_result):
    print("plot_comparison_result_impl")

    xr_curve = corrected_sd.get_xr_curve()
    x = xr_curve.x
    y = xr_curve.y

    sg_array, at_array, al_array = comparison_result
    in_folder = get_in_folder()

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Comparison of Rg values of %s" % in_folder, fontsize=16)
        ax.plot(x, y)
        axt = ax.twinx()
        axt.grid(False)
        axt.plot(x, sg_array[:,0], color="cyan", label="_MOLASS")
        axt.plot(x, at_array[:,0], color="red", label="ATSAS")
        axt.plot(x, al_array[:,0], color="green", label="Alsaker")
        axt.legend()
        fig.tight_layout()
        plt.show()    


# coding: utf-8
"""
    Baseline.Demo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
from matplotlib.patches import Polygon
from MeasuredData import MeasuredData
import molass_legacy.KekLib.DebugPlot as plt
from ThreeDimViewer import ThreeDimViewer
from MatrixData import simple_plot_3d
from .Baseline import compute_baseline, better_integrative_curve
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from CurveDecomposer import decompose

BUG_FIX_WITH_REDRAW = False
TITLE_FONTSIZE = 16

def demo(root, in_folder, xray=True, show_preview=False):
    print(in_folder)
    md = MeasuredData(in_folder)

    if show_preview:
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    if xray:
        rd = md.xr
        ecolor = 'orange'
    else:
        rd = md.uv
        ecolor = 'blue'

    rd.set_elution_curve()
    i_slice = rd.i_slice
    j_slice = rd.j_slice

    whole_data = rd.data
    data = whole_data[i_slice,j_slice]
    q = rd.vector[i_slice]

    fig = plt.figure(figsize=(21, 11))
    gs = GridSpec(2,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1])
    ax02 = fig.add_subplot(gs[0,2])
    ax10 = fig.add_subplot(gs[1,0])
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])

    ax00.set_title("3D Data View", y=1.09, fontsize=TITLE_FONTSIZE)
    simple_plot_3d(ax00, data, x=q)
    z = rd.e_curve.y[j_slice]

    q1 = rd.vector[rd.e_index]
    size = data.shape[1]
    x = np.ones(size)*q1
    y = np.arange(size)

    ax00.plot(x, y, z, color=ecolor)

    baseline = None
    for ax1, ax2, integral in [
            (ax01, ax02, False), 
            (ax11, ax12, True),
            ]:
        if integral:
            title = "LPM Integral Baseline"
        else:
            title = "LPM Linear Baseline"
        b = compute_baseline(z, integral=integral)
        if baseline is None:
            baseline = b
        z_ = z - b
        ec = ElutionCurve(z_)
        recs = decompose(ec)

        ax1.set_title(title, fontsize=TITLE_FONTSIZE)
        ax1.plot(z, color=ecolor)
        ax1.plot(b, ':', color='red')
        ax2.set_title("Decomposition after Correction", fontsize=TITLE_FONTSIZE)
        ax2.plot(z_, color=ecolor)
        for k, rec in enumerate(recs):
            func = rec.evaluator
            ax2.plot(y, func(y), ':', color='C%d' % k)

    # ax10.set_ylim(-0.01, 0.2)

    anim = make_anim(fig, ax10, y, z, baseline, ecolor)
    fig.tight_layout()
    # anim.save("anim.mp4", writer="ffmpeg")
    plt.show()

def draw_anim_init(ax, x, y, b, ecolor, n):
    ax.set_title("LPM Integral Baseline Algorithm Animation", fontsize=TITLE_FONTSIZE)
    ax.plot(x, y, color=ecolor)
    line0, = ax.plot(x, b, ':', color='green')
    area_points = [(x[0], b[0])] + list(zip(x,y)) + [(x[-1], b[0])]
    area1 = Polygon(area_points, alpha=0.3, color='cyan')
    ax.add_patch(area1)

    tx = np.average(ax.get_xlim())
    ymin, ymax = ax.get_ylim()
    ty = ymin*0.6 + ymax*0.4
    num_text = ax.text(tx, ty, str(n), alpha=0.05, fontsize=300, ha="center", va="center")
    return line0, area1, num_text

def make_anim(fig, ax, x, y, b, ecolor):

    line0, area1, num_text = draw_anim_init(ax, x, y, b, ecolor, 1)

    num_frames1 = 60
    num_frames2 = 60
    num_frames3 = 30
    num_frames4 = 60
    num_frames = num_frames1 + num_frames2 + num_frames3 + num_frames4

    ymin, ymax = ax.get_ylim()

    smear_rate = 0.05
    end_height = b[-1]  # or y[-1]
    area2 = None
    integ_height = ymax*0.2

    start_y = y[0]
    y_ = y - b
    cy = np.cumsum(y_*smear_rate)
    height = cy[-1] - cy[0]
    ratio = (y[-1] - start_y)/height
    integ_y = cy*ratio

    b2 = start_y + integ_y
    y2_ = y - b2
    cy2 = np.cumsum(y2_)
    height = cy2[-1] - cy2[0]
    ratio = y2_[-1]/height
    integ_y2 = cy2*ratio

    lineb = None

    def make_area_points(y1, y2):
        r = list(zip(x,y2))
        r.reverse()
        return list(zip(x,y1)) + r

    def init():
        line0.set_data(x, b)
        num_text.set_text("1")
        return line0, area1, num_text

    def animate_impl(i, a_integ_y, a_baseline):
        nonlocal area2, lineb
        if i < num_frames1:
            ratio = smear_rate + (1 - smear_rate)*(1 - i/num_frames1)
            # print([i], ratio)
            yp = a_baseline + y_*ratio
            area_points = make_area_points(yp, a_baseline)
            area1.set_xy(area_points)
            return line0, area1, num_text

        if i < num_frames1 + num_frames2:
            i_ = i - num_frames1
            if i_ == 0:
                area_points = [(x[0], start_y), (x[0], start_y)]
                area2 = Polygon(area_points, alpha=0.3, color='pink')
                ax.add_patch(area2)
            else:
                # i.e. i_ > 0
                j = int(len(x)*i_/num_frames2)
                yp = a_baseline[0] + a_integ_y[0:j]
                area_points = [(x[0], start_y)] + list(zip(x[0:j], yp)) + [(x[max(0,j-1)], start_y)]
                area2.set_xy(area_points)
            return line0, area1, area2, num_text

        if i < num_frames1 + num_frames2 + num_frames3:
            i_ = i - (num_frames1 + num_frames2)
            w = (i_+1)/num_frames3
            yp = a_baseline[0] + a_integ_y*( 1*(1-w) + (a_baseline[-1]-a_baseline[0])/a_integ_y[-1]*w )
            area_points = [(x[0], start_y)] + list(zip(x, yp)) + [(x[-1], start_y)]
            area2.set_xy(area_points)
            return line0, area1, area2, num_text

        i_ = i - (num_frames1 + num_frames2 + num_frames3)
        if i_ == 0:
            lineb, = ax.plot(x, b2, ':', color='red')

        return line0, area1, area2, lineb, num_text

    def animate(i):
        nonlocal line0, area1, num_text

        if i < num_frames:
            return animate_impl(i, integ_y*integ_height/integ_y[-1], b)

        i -= num_frames
        if i == 0:
            if BUG_FIX_WITH_REDRAW:
                ax.cla()
                line0, area1, num_text = draw_anim_init(ax, x, y, b2, ecolor, 2)
            else:
                line0.set_data(x, b2)
                area2.remove()
                num_text.set_text("2")

        return animate_impl(i, integ_y2*integ_height/integ_y2[-1], b2)

    anim = animation.FuncAnimation(fig, animate, init_func=init,
                               frames=num_frames*2, interval=100, blit=True)
    return anim

def proof_plot(in_folder, xray=True, better=True):
    from DataUtils import get_in_folder

    md = MeasuredData(in_folder)

    if xray:
        rd = md.xr
        ecolor = 'orange'
    else:
        rd = md.uv
        ecolor = 'blue'

    j_slice = rd.j_slice
    x = rd.e_curve.x[j_slice]
    y = rd.e_curve.y[j_slice]

    end_slices = rd.e_curve.get_end_slices()
    if better:
        b, convex = better_integrative_curve(y, end_slices=end_slices)
    else:
        b = compute_baseline(y, integral=True, end_slices=end_slices)

    fig, ax = plt.subplots()
    ax.set_title("Proof Plot of Integral Basecurve for " + get_in_folder(in_folder), fontsize=20)
    ax.plot(x, y, color=ecolor, label='data')
    ax.plot(x, b, ':', color='red', label='integral basecurve')

    ax.legend(fontsize=16)
    fig.tight_layout()
    plt.show()



# coding: utf-8
"""
    ElutionModelEditor.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import os
import copy
import numpy                as np
import matplotlib.pyplot    as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar, get_color, get_hex_color
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
import tkinter.font as font
from molass_legacy.KekLib.TkSupplements          import set_icon
from TkMiniTable            import TkMiniTable
from molass_legacy.Models.ElutionCurveModels     import EGHA, EMGA
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting

class ElutionModelEditor(Dialog):
    def __init__(self, parent, x, y, opt_recs):
        self.parent = parent
        self.x = x
        self.y = y
        self.opt_recs = opt_recs
        self.num_eltns = len(self.opt_recs)
        Dialog.__init__( self, parent, "Elutio nModel Editor", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame): # overrides parent class method
        set_icon(self)

        canvas_frame = Tk.Frame( body_frame )
        canvas_frame.pack(side=Tk.LEFT)
        editor_frame = Tk.Frame( body_frame )
        editor_frame.pack(side=Tk.LEFT)

        self.fig = plt.figure( figsize=(7, 6) )
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, canvas_frame )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.ax = self.fig.gca()
        self.draw()
        self.fig.tight_layout()

        self.create_table(editor_frame)

    def draw(self):
        ax = self.ax
        x = self.x
        y = self.y
        ax.cla()
        ax.plot(x, y, color='orange', label='data')
        total_y = np.zeros(len(y))
        resid_y = copy.deepcopy(y)
        for k, rec in enumerate(self.opt_recs):
            func = rec[1]
            y_ = func(x)
            ax.plot(x, y_, ':', label='component %d' % (k+1), color=get_color(k), linewidth=3)
            total_y += y_
            resid_y -= y_

        ax.plot(x, total_y, ':', label='model total', color='red', linewidth=3)
        ax.plot(x, resid_y, label='residual', color='gray')
        ax.legend()
        self.mpl_canvas.draw()

    def create_table(self, frame):
        table = TkMiniTable( frame,
            columns = [ '_', 'A', 'B', 'C', 'D' ],
            default_colwidth = 20,
            # font = font.Font(self.parent, family="Ariel", size=16),
            font = ("", 12),
            cell_widget_class=Tk.Entry,
            )

        table.heading( '_', text='No' )
        table.column ( '_', justify=Tk.CENTER, width=5 )
        table.heading( 'A', text='Color' )
        table.heading( 'B', text='Type' )
        table.heading( 'C', text='Definition' )
        table.heading( 'D', text='Hide' )
        table.column ( [1], width=10 )
        table.column ( [2], width=10 )
        table.column ( [3], width=60 )
        table.column ( [4], width= 5 )

        table.insert_row( ( 0, "", "Data", "", "" ) )

        for k, rec in enumerate(self.opt_recs):
            evaluator = rec[1]
            type_ = evaluator.get_model_name()
            def_ = evaluator.get_model_def_expr()
            cells = table.insert_row( ( 1+k, "", type_, def_, "" ) )
            for c in cells[2:]:
                c.widget.config(bg='honeydew')

        table.insert_row( ( self.num_eltns+1, "", "Total", "", "" ) )
        table.insert_row( ( self.num_eltns+2, "", "Residual", "", "" ) )

        table.pack()



"""
    Models.MomentFitting.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Elution.CurveUtils import simple_plot
from molass_legacy.Peaks.ElutionModels import compute_moments
from molass_legacy.Models.Stochastic.SDM import SDM

def sdm_guess_impl(model, x, y, moments):
    def objective(p):
        pass

def fit_a_component(model, x, y):
    M = compute_moments(x, y)
    sdm_guess_impl(model, x, y, M)

def spike(caller):
    model = SDM()
    sd = caller.serial_data
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xr_curve()
    x = xr_curve.x
    y = xr_curve.y
    fit_a_component(model, x, y)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("MomentFitting Spike", fontsize=20)
        simple_plot(ax1, uv_curve, color="blue")
        simple_plot(ax2, xr_curve, color="orange")
        fig.tight_layout()
        plt.show()


# coding: utf-8
"""
    EFA.RotationDemo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Polygon
import matplotlib.animation as animation
from CorrectedData import CorrectedXray
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.OurMatplotlib import get_color
from MatrixData import simple_plot_3d
from ModeledData import ModeledData
from Tutorials.MatrixFactorization import draw_PQ, generate_random_points
from Tutorials.SvdPrecision import plot_text

FONTSIZE = 200
A_SIZE = 600

def model_demo(bogus_init=False, random_init=False, symmetric=True, save_anim=False):
    fig = plt.figure(figsize=(21,11))

    proof_type = 'Proof' if symmetric else 'Weakness Proof'
    fig.suptitle("Iterative Rotation Method %s" % proof_type, fontsize=40)

    pause = False

    def on_click(event):
        nonlocal pause
        print('on_click')
        if event.inaxes is None:
            return
        pause ^= True

    plt.dp.mpl_canvas.mpl_connect('button_press_event', on_click)

    q = np.linspace(0.01, 0.6, A_SIZE)
    rg_list = [50, 24]
    d_list = [1, 3]
    if symmetric:
        E_SIZE = 300
        md = ModeledData(q, E_SIZE, rg_list=rg_list, d_list=d_list)
        f1, t1 = 0, 200
        f2, t2 = 100, E_SIZE
        num_frames = 20
        interval = 1000
    else:
        E_SIZE = 500
        md = ModeledData(q, E_SIZE, rg_list=rg_list, d_list=d_list,
                            h_list=[1, 0.5], mu_list=[130, 220], sigma_list=[20, 30], tau_list=[40, 20])
        f1, t1 = 0, 350
        f2, t2 = 120, E_SIZE
        num_frames = 1000
        interval = 50

    if False:
        dplt.push()
        md.plot_components()
        dplt.pop()
    M = md.get_data()

    rank = len(rg_list)

    U, s, VT = np.linalg.svd(M)

    Us_ = np.dot( U[:,0:rank], np.diag( s[0:rank] ) )
    M_  = np.dot( Us_, VT[0:rank,:] )

    gs = GridSpec(2,4)
    axes = []
    for i in range(2):
        arow = []
        for k in range(4):
            proj = '3d' if i==0 and k == 0 else None
            arow.append(fig.add_subplot(gs[i,k], projection=proj))
        axes.append(arow)
    ax1, ax2, ax3, ax4  = axes[0]
    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    for i in range(rank):
        ax2.plot(q, U[:,i])
        ax4.plot(VT[i,:])

    simple_plot_3d(ax1, M_, x=q)
    ns = 5
    ax3.plot(s[0:ns], ':', color='gray')
    for k in range(ns):
        ax3.plot(k, s[k], 'o', color=get_color(k))

    plot_text(ax2, "$U$", fontsize=FONTSIZE)
    plot_text(ax3, "$\Sigma$", fontsize=FONTSIZE)
    plot_text(ax4, "$V^T$", fontsize=FONTSIZE)

    ax5, ax6, ax7, ax8  = axes[1]

    n = md.num_components
    for k in range(n):
        ax5.plot(md.i, md.P[:,k])
        ax6.plot(md.C[k,:])

    poly1 = Polygon([(f1,0), (f1,1), (t1,1), (t1,0)], color='cyan', alpha=0.2)
    ax6.add_patch(poly1)
    poly2 = Polygon([(f2,0), (f2,1), (t2,1), (t2,0)], color='pink', alpha=0.2)
    ax6.add_patch(poly2)

    def make_C():
        if bogus_init:
            if random_init:
                c1 = np.random.uniform(0,1, E_SIZE)
                c2 = 1 - c1
            else:
                c1 = np.zeros(E_SIZE)
                c1[f1:f2] = 1
                c1[f2:t1] = 0.5
                c2 = np.zeros(E_SIZE)
                c2[f2:t1] = 0.5
                c2[t1:t2] = 1
            C = np.vstack([c1, c2])
        else:
            C = VT[0:rank,:].copy()
        return C

    C = make_C()
    R = np.zeros(C.shape)
    R[0,f1:t1] = 1
    R[1,f2:t2] = 1

    if bogus_init:
        ax4.plot(C[0,:], ':')
        ax4.plot(C[1,:], ':')

    x = np.arange(E_SIZE)

    def draw_C_lines(C):
        line_c1, = ax8.plot(x, C[0,:])
        line_c2, = ax8.plot(x, C[1,:])
        return line_c1, line_c2

    line_c1, line_c2 = draw_C_lines(C)

    P = np.dot(M_, np.linalg.pinv(C))

    def draw_P_lines(P):
        line_p1, = ax7.plot(q, P[:,0])
        line_p2, = ax7.plot(q, P[:,1])
        return line_p1, line_p2

    line_p1, line_p2 = draw_P_lines(P)

    def clear_axes():
        for ax in [ax7, ax8]:
            ax.cla()

    def draw_texts(text):
        ret = []
        for ax in [ax7, ax8]:
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            t = ax.text(tx, ty, text, ha="center", va="center", alpha=0.1, fontsize=200)
            ret.append(t)
        return ret

    text7, text8 = draw_texts("0")

    blit = False        # it is difficult to update axis labels with blit=True

    def update_text(text):
        for ax, t in [(ax7, text7), (ax8, text8)]:
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            t.set_position((tx, ty))
            t.set_text(text)

    artists = [line_c1, line_c2, line_p1, line_p2, text7, text8]

    def update_C_lines(C):
        line_c1.set_data(x,C[0,:])
        line_c2.set_data(x,C[1,:])

    def update_P_lines(P):
        line_p1.set_data(q,P[:,0])
        line_p2.set_data(q,P[:,1])

    def reset():
        nonlocal P, C
        C = make_C()
        P = np.dot(M_, np.linalg.pinv(C))

        if blit:
            update_C_lines(C)
            update_P_lines(P)
            update_text("0")
            return artists
        else:
            clear_axes
            draw_C_lines(C)
            draw_P_lines(P)
            draw_texts("0")

    def normalize_C(C):
        C = R*C
        for k in range(C.shape[0]):
            C[k,:] /= np.sum(C[k,:])
        return C

    def animate(i):
        nonlocal P, C
        print([i])
        if i % 2 == 0:
            P = np.dot(M_, np.linalg.pinv(C))
            C = np.dot(np.linalg.pinv(P), M_)
        else:
            C = normalize_C(C)

        num_text = str(i//2)

        if blit:
            update_C_lines(C)
            update_P_lines(P)
            ax7.set_ylim(P.min(), P.max())
            ax8.set_ylim(C.min(), C.max())
            update_text(num_text)
            return artists
        else:
            clear_axes()
            draw_C_lines(C)
            draw_P_lines(P)
            draw_texts(num_text)

    def index_generator():
        i = 0
        while i < num_frames:
            i_ = i
            if pause:
                pass
            else:
                i += 1
            yield i_

    save_count = 1500 if save_anim else None
    anim = animation.FuncAnimation(fig, animate, index_generator, blit=blit, init_func=reset, interval=interval
                                    , save_count=save_count)
    if save_anim:
        anim.save("anim.mp4", writer="ffmpeg")

    plt.show()



# coding: utf-8
"""
    BpoDependencyDemo.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.stats import norm
from bisect import bisect_right
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FuncFormatter
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

N  = 301
NUM_ITERATIONS  = 100
MIN_SIGMA = 3
MAX_SIGMA = 7

class MyFormatter:
    def __init__(self, x):
        self.x  = x

    def x_formatter( self, val, pos):
        if pos is None: return ''
        # print( 'val=', val )
        ival = int(val)
        if ival >= 0 and ival <= N:
            return '%g\n(%.2gσ)' % ( val, self.x[ival] )
        else:
            return '%g' % ( val )

def demo_func(fig):
    ym  = norm.pdf( 0, 0, 1 )

    rec_list = []
    params_list = []

    for w in np.linspace( MIN_SIGMA, MAX_SIGMA, 21 ):

        x   = np.linspace(-w, w, N )

        if w == 3:
            i   = 0
            wi  = w
            xi  = x
        elif abs( w - 5 ) < 1e-5:
            i   = 1
            wi  = w
            xi  = x
        elif w == 7:
            i   = 2
            wi  = w
            xi  = x
        else:
            i   = None
            wi  = None
            xi  = None

        y0  = norm.pdf( x, 0, 1 )
        scale = 1/ym
        y_  = y0 * scale

        x_array = []
        p_array = []
        s_array = []

        params_row = []
        for n in range(50):
            if n == 0:
                j = 0
            elif n == 24:
                j = 1
            elif n == 49:
                j = 2
            else:
                j = None

            nn = 0.01 * (n+1)
            ys_array = []
            zi_array = []
            yi  = None
            yi_ = None
            for k in range(NUM_ITERATIONS):
                noise = np.random.normal(0, nn, N)
                y   = y_ + noise
                ys  = sorted( y )
                zi  = bisect_right( ys, 0 )
                zi_array.append( zi )
                ys_array.append( ys[zi] )
                if k == 0  and i is not None and j is not None:
                    yi_ = y_
                    yi  = y

            p = np.average( zi_array ) * 100 / N
            x_array.append( nn )
            p_array.append( p )
            s_array.append( np.std( zi_array ) )

            if i is not None and j is not None:
                params_row.append( [ wi, nn, p ] )

        rec_list.append( [ w, x_array, p_array, s_array ] )
        if i is not None:
            params_list.append( params_row )

    gs = gridspec.GridSpec( 3, 6 )
    ax3d  = fig.add_subplot( gs[:,0:3], projection='3d' )
    ax3d.set_title( 'Noise and Sigma-width Dependency of Base Percentile Offset' )
    ax3d.set_xlabel( 'noise' )
    ax3d.set_ylabel( 'width(σ)' )
    ax3d.set_zlabel( 'base percentile offset' )

    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors = prop_cycle.by_key()['color']
    # print( 'colors=', colors )

    for rec in rec_list:
        w, x_array, p_array, s_array = rec
        y_array = np.ones( len(x_array) ) * w
        ax3d.plot( x_array, y_array, s_array, color='yellow' )
        ax3d.plot( x_array, y_array, p_array, color='orange' )

    params = np.array( params_list )


    for i in range(3):
        for j in range(3):
            wi, nn, p = params[i,j]
            x   = np.linspace(-wi, wi, N )
            y0  = norm.pdf( x, 0, 1 )
            scale = 1/ym
            y_  = y0 * scale
            noise = np.random.normal(0, nn, N)
            y   = y_ + noise

            sy  = sorted( y )
            m   = bisect_right( sy, nn )
            pn  = m/N
            print( (i, j), pn )

            ax  = fig.add_subplot( gs[i,3+j] )
            ax.set_ylim(-1, 2)

            fmt = MyFormatter(x)
            formatter   = FuncFormatter( fmt.x_formatter )
            ax.xaxis.set_major_formatter( formatter )
            # TODO: pn ?
            # title = 'width=%g; noise=%g; bpo=%.2g; %.2g' % ( wi, nn, p, pn*0.5 )
            title = 'width=%g; noise=%g; bpo=%.2g' % ( wi, nn, p )
            ax.set_title( title )
            ax.plot( y )
            ax.plot( y_ )
            ax.plot( [0, 300], [0, 0], ':', color='red' )
            # ax.plot( x, y )
            # ax.plot( x, y_ )
            fwhm = 2 * np.sqrt( 2*np.log(2) ) * 150/wi
            print( 'fwhm[%d,%d]=%.3g' % (i, j, fwhm) )

            ax3d.scatter( nn, wi, p, color='red' )

    fig.tight_layout()

class BpoDependencyDemo(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.log_scale = 1
        Dialog.__init__(self, parent, "Noise and Sigma-width Dependency of BPO", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X)
        tbframe = Tk.Frame(bframe)
        tbframe.pack(side=Tk.LEFT)

        self.fig = plt.figure( figsize=(18, 9)  )
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        demo_func(self.fig)
        self.fig.tight_layout()
        self.mpl_canvas.draw()

        # self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe, show_mode=False )
        self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe )
        self.toolbar.update()

    def show(self):
        self._show()



"""
    Models.Stochastic.DispersivePdfStudy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.widgets import Slider, Button
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, MAX_PORESIZE
from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf

def study_pdf():
    N = 2000
    T = 0.5
    K = N*T
    me = 1.5
    mp = 1.5
    poresize = 75
    rgs = np.array([50, 40, 30, 20])
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    t0 = 500
    N0 = 14400
    t = np.arange(0, 1000)

    def compute_curves(x, N, T, x0, tI, N0, rhov):
        cy_list = []
        for rho in rhov:
            ni = N*(1 - rho)**me
            ti = T*(1 - rho)**mp
            cy = dispersive_monopore_pdf(x - tI, ni, ti, N0, x0 - tI)
            cy_list.append(cy)
        return cy_list

    rv = np.linspace(1, 70, 100)
    def compute_trv(N, T, poresize):
        rhov_ = rv/poresize
        rhov_[rhov_ > 1] = 1
        niv = N*(1 - rhov_)**me
        tiv = T*(1 - rhov_)**mp
        trv = t0 + niv*tiv
        return trv
    trv = compute_trv(N, T, poresize)

    def plot_curves(ax, x, x0, tI, title, with_arrow=False):
        ax.set_title(title + " with $t_{I}$=%g" % tI, fontsize=16)
        cy_list = compute_curves(x, N, T, x0, tI, N0, rhov)
        curves = []
        for k, (cy, rg) in enumerate(zip(cy_list, rgs)):
            curve, = ax.plot(x, cy, ':', label='$R_g$=%g' % rg)
            curves.append(curve)
        lines = []
        line = ax.axvline(x=x0, color='red', label='$t_0$')
        lines.append(line)
        line = ax.axvline(x=tI, color='gray', label='$t_I$')
        lines.append(line)
        if with_arrow:
            hy = np.average(ax.get_ylim())
            dx = x0 - tI
            ax.arrow(x=tI, y=hy, dx=dx, dy=0, width=0.005, head_width=0.015,
                    head_length=0.2*dx, length_includes_head=True, color='pink')
        ax.legend()

        axt = ax.twinx()
        axt.grid(False)
        excl_curve, = axt.plot(trv + tI, rv, color='yellow', label='exclusion curve')
        axt.legend(loc="center right")

        return curves, lines, excl_curve

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(18,5))
        fig.suptitle("SDM PDF Study in Experimental Time Axis with Poresize=%g, N*T=%g" % (poresize, K), fontsize=20)

        curves1, lines1, excl_curve1 = plot_curves(ax1, t, t0, 0, "Theoretical Time Axis")

        tI = -500
        x = t + tI
        x0 = t0 + tI
        curves2, lines2, excl_curve2 = plot_curves(ax2, x, x0, tI, "Experimental Time Axis")

        fig.tight_layout()
        fig.subplots_adjust(right=0.7, wspace=0.2)

        slider_params = [N, K, t0, poresize, N0, tI]
        bounds = [(100, 3000), (100, 3000), (0, 1000), (30, MAX_PORESIZE), (100, 30000), (-1000, 0)]

        slider_specs = [    (r"$N$", *bounds[0], slider_params[0]),
                            (r"$K$", *bounds[1], slider_params[1]),
                            (r"$t_0$", *bounds[2], slider_params[2]),
                            (r"$poresize$", *bounds[3], slider_params[3]),
                            (r"$N_0$", *bounds[4], slider_params[4]),
                            (r"$t_I$", *bounds[5], slider_params[5]),
                            ]

        def slider_update(k, val):
            # print([k], "slider_update", val)
            slider_params[k] = val
            N_, K_, t0_, poresize_, N0_, tI_ = slider_params
            T_ = K_/N_
            rhov_ = rgs/poresize_
            cy_list1 = compute_curves(t, N_, T_, t0_, 0, N0_, rhov_)
            for curve, cy in zip(curves1, cy_list1):
                curve.set_data(t, cy)
            for line, pos in zip(lines1, [t0_, 0]):
                line.set_xdata(pos)

            x = t + tI_
            x0 = t0_ + tI_
            cy_list2 = compute_curves(x, N_, T_, x0, tI_, N0_, rhov_)
            for curve, cy in zip(curves2, cy_list2):
                curve.set_data(x, cy)
            for line, pos in zip(lines2, [x0, tI_]):
                line.set_xdata(pos)

            trv_ = compute_trv(N_, T_, poresize_)
            excl_curve1.set_data(trv_, rv)
            excl_curve2.set_data(trv_ + tI_, rv)

            fig.canvas.draw_idle()

        left = 0.78
        width = 0.15
        slider_axes = []
        sliders = []

        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([left, 0.8 - 0.08*k, width, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = fig.add_axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        ret = plt.show()
    return ret


"""
    LoaderProcess.DataLoader.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time, sleep
import glob
import numpy as np

class DataLoader:
    def __init__(self, si, rg_queue):
        self.si = si
        self.rg_queue = rg_queue

    def load(self, in_folder, notify=True):
        t0 = time()
        paths = glob.glob(in_folder + '/*.dat')

        for i, path in enumerate(paths):
            print([i], path)
            d = np.loadtxt(path)
            if i == 0:
                shape = (len(paths), *d.shape)
                self.si.initilize_buffer(shape)
                if notify:
                    self.rg_queue.put(-1)   # notify to get ready for the folder
            self.si.set_xr_data(i, d)
            if notify:
                self.rg_queue.put(i)

        if notify:
            self.rg_queue.put(-2)   # notify to get finished for the folder
        print("It took %.3g seconds" % (time()-t0))



"""
    Sympy/StochasticUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

from sympy import symbols, exp, diff, simplify, I

w = symbols('w')

def raw_moment(cf, k):
    return simplify((-I)**k * diff(cf, w, k).subs(dict(w=0)))

def central_moment(cf, k):
    m = raw_moment(cf, 1)
    return simplify(raw_moment(cf/exp(I*w*m),k))

def demo_monopore_cf():
    n1, t1 = symbols('n1, t1')
    cf1 = exp(n1*(1/(1 - I*w*t1) - 1))
    print("raw_moment(cf1, 1)=", raw_moment(cf1, 1))
    print("central_moment(cf1, 2)=", central_moment(cf1, 2))
    print("central_moment(cf2, 3)=", central_moment(cf1, 3))

def demo_dipore_cf():
    n1, t1, n2, t2 = symbols('n1, t1, n2, t2')
    cf2 = exp(n1*(1/(1 - I*w*t1) - 1)) * exp(n2*(1/(1 - I*w*t2) - 1))
    print("raw_moment(cf2, 1)=", raw_moment(cf2, 1))
    print("central_moment(cf2, 2)=", central_moment(cf2, 2))
    print("central_moment(cf2, 3)=", central_moment(cf2, 3))

if __name__ == "__main__":
    # demo_monopore_cf()
    demo_dipore_cf()


"""
    Optimizer.UvBaselineEstimator.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import basinhopping
LOG_IDENTIFICATION = "FixedBaseline Optimizer"

class UvBaselineEstimator:
    def __init__(self, optimizer):
        self.logger = logging.getLogger(__name__)
        self.optimizer = optimizer
        self.logger.info("%s constructed", LOG_IDENTIFICATION)

    def prepare(self, init_params, real_bounds=None):
        self.optimizer.prepare_for_optimization(init_params, real_bounds=real_bounds)
        nc = self.optimizer.num_pure_components

        model_name = self.optimizer.get_model_name()

        if model_name == "EGH":
            xr_start = nc*5
        elif model_name == "SDM":
            xr_start = nc
        elif model_name == "EDM":
            xr_start = nc*7
        else:
            assert False

        xr_stop = xr_start + 2
        uv_start = xr_stop + nc
        uv_indeces = np.arange(uv_start, uv_start+7)
        self.indeces = uv_indeces
        self.logger.info("%s prepared with model=%s, indeces=%s", LOG_IDENTIFICATION, model_name, str(list(self.indeces)))

    def solve(self, init_params, real_bounds=None, niter=100, seed=None, open_mode="w", debug=False):
        self.logger.info("FixedBaseline solve started with init_params=%s", str(init_params))
        self.prepare(init_params, real_bounds=real_bounds)

        self.cb_fh = open("callback.txt", open_mode)
        self.optimizer.cb_fh = self.cb_fh
        self.optimizer.write_init_callback_txt(init_params)

        self.temp_params = self.optimizer.to_norm_params(init_params)

        bounds = [(0, 10)]*len(self.indeces)
        minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds)
        result = basinhopping(self.objective_func_wrapper, self.temp_params[self.indeces],
                              callback=self.minima_callback,
                              niter=niter,
                              seed=seed,
                              minimizer_kwargs=minimizer_kwargs)

        self.cb_fh.close()
        self.temp_params[self.indeces] = result.x
        result.x = self.temp_params
        return result
    
    def objective_func_wrapper(self, p):
        self.temp_params[self.indeces] = p
        real_params = self.optimizer.to_real_params(self.temp_params)
        return self.optimizer.objective_func(real_params)
    
    def minima_callback(self, x, f, accept):
        self.temp_params[self.indeces] = x
        return self.optimizer.minima_callback(self.temp_params, f, accept)

def test_estimate_uvbaseline_impl(optimizer, init_params):
    print("test_estimate_baseline_impl")
    optimizer.objective_func(init_params, plot=True)
    estimator = UvBaselineEstimator(optimizer)
    result = estimator.solve(init_params)
    optimizer.objective_func(result.x, plot=True)


#!/usr/bin/env python
from numpy import *

def generate_hex_circle_packing(a, width):
    """Generate a domain of a given width filled with hexagonally packed
    circles.  The height will be determined so that the vertical 
    boundary condition is periodic.

    Arguments:
    a       particle radius
    width   domain width, in terms of particle radius

    Returns:
    x_list  list of x coordinates
    y_list  list of y coordinates
    x_size  width of domain (equal to argument width)
    y_size  height of domain
    """
    numParticles = 0

    x_list = []
    y_list = []
    y = a
    x = a
    rowNumber = 0
    # Create a row
    while y <= width*1.01:
        # Create circles in a row
        while x < width:
            x_list.append(x)
            x = x + 2*a
            y_list.append(y)
            numParticles = numParticles + 1
        y = y + a*sqrt(3.)
        rowNumber = rowNumber + 1
        if rowNumber%2 == 0:
            x = a
        else:
            x = 0
    x_size = width
    y_size = rowNumber*a*sqrt(3)

    return array(x_list), array(y_list), x_size, y_size

def plot_adsorbed_circles(adsorbed_x, adsorbed_y, radius, width, reference_indices=[]):
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.patches import Circle

    # Plot each run
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for p in range(len(adsorbed_x)):
        if len(np.where(reference_indices == p)[0]) > 0:
            ax.add_patch(Circle((adsorbed_x[p], adsorbed_y[p]), radius, 
                edgecolor='black', facecolor='black'))
        else:
            ax.add_patch(Circle((adsorbed_x[p], adsorbed_y[p]), radius,
                edgecolor='black', facecolor='white'))

    ax.set_aspect(1.0)
    plt.axhline(y=0, color='k')
    plt.axhline(y=width, color='k')
    plt.axvline(x=0, color='k')
    plt.axvline(x=width, color='k')
    plt.axis([-0.1*width, width*1.1, -0.1*width, width*1.1])
    plt.xlabel("non-dimensional x")
    plt.ylabel("non-dimensional y")

    return ax




"""
    Optimizer.RgCurveInspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from .RatioInterpretIllust import MODEL_NAME_DICT
from .FvScoreConverter import convert_score

def rg_curve_inspect(caller):
    from importlib import reload
    import Optimizer.FuncReloadUtils
    reload(Optimizer.FuncReloadUtils)
    from molass_legacy.Optimizer.FuncReloadUtils import reload_optimizer

    print("rg-curve inspect")
    js_canvas = caller.canvas
    optimizer = reload_optimizer(js_canvas.fullopt)
    params = js_canvas.get_current_params()
    lrf_info = optimizer.objective_func(params, return_lrf_info=True, debug=True)   # this sets optimizer.separate_params to update rg_params

    optimizer.update_guinier_region(params=params, debug=True)     # in order to reload GuinierDeviation, which is currently under revision

    job_name = js_canvas.dialog.get_job_info()[0]
    in_folder = get_in_folder()

    Pxr = lrf_info.matrices[0]
    Cxr = lrf_info.matrices[1]
    rg_params = optimizer.separate_params[2]
    
    Guinier_deviation = optimizer.get_guinier_deviation(Pxr, Cxr, rg_params, debug=True)
    if Guinier_deviation is None:
        return
    print("Guinier_deviation=%.3g" % Guinier_deviation)

    fv = optimizer.objective_func(params)
    sv = convert_score(fv)

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18, 5))
        ax1.set_title("UV Decomoposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)
        ax3.set_title("Objecive Function Scores in SV=%.3g" % sv, fontsize=16)
        axt = ax2.twinx()
        axt.grid(False)

        optimizer.objective_func(params, plot=True, axis_info=(fig, (ax1, ax2, ax3, axt)))

        fig.tight_layout()
        plt.show()


"""
    SecTheory.SecSimulator.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""

import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Slider, Button
from matplotlib.patches import Rectangle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Models.ElutionCurveModels import egh
from SaferSpinbox import SaferSpinbox
from SecTheory.MwRgFigure import get_mwrg_info

class SecSimulator(Dialog):
    def __init__(self, parent, optimizer=None, params=None):
        self.optimizer = optimizer
        self.params = params
        if optimizer is None:
            self.initilize_for_unit_test()
        else:
            self.initialize_from_optimizer(optimizer, params)

        self.lrf_info = None
        self.gk_info = None

        Dialog.__init__(self, parent, "SEC Simulator", visible=False)

    def initilize_for_unit_test(self):
        rp = 87.7
        tI = -100
        t0 = 100
        P = 1000
        m = 3.0
        self.set_slider_specs(rp, tI, t0, P, m)

        self.init_rgs = [35.7, 35.0]
        self.init_prs = [80, 20]
        self.init_tss = [1.0, 1.0]
        self.num_plates_disclosed = 48000

    def initialize_from_optimizer(self, optimizer, params):
        self.separate_params = optimizer.split_as_unified_params(params)
        xr_params = self.separate_params[0]
        rg_params = self.separate_params[2]
        seccol_params = self.separate_params[7]
        Npc, rp, tI, t0, P, m = seccol_params
        Np = round(Npc/0.3)

        self.set_slider_specs(rp, tI, t0, P, m)

        x = optimizer.xr_curve.x

        opt_lrf_info = optimizer.objective_func(params, return_lrf_info=True)
        proportions = opt_lrf_info.get_xr_proportions()

        self.init_rgs = rg_params
        self.init_prs = proportions*100
        self.init_tss = xr_params[:,3]/xr_params[:,2]
        self.num_plates_disclosed = round(Np)

    def set_slider_specs(self, rp, tI, t0, P, m):
        self.init_slider_specs = [
            ("$r_p$",           60,         350,    rp),
            ("$t_I$",     tI - 500,    tI + 300,    tI),
            ("$t_0$",     t0 - 1000,    t0 + 300,    t0),
            ("$P$",       1000, 10000,    P),
            ("$m$",       1,      6,     m),
            ]
        self.rp_init = rp
        self.excl_limit_init = get_setting('exclusion_limit')

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        xframe = Tk.Frame(body_frame, bg='white')
        xframe.place(relx=0.74, rely=0.08)
        rframe = Tk.Frame(body_frame, bg='white')
        rframe.place(relx=0.72, rely=0.58)
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, padx=10)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)

        self.fig = fig = plt.figure(figsize=(18,8))
        self.create_axes()
        self.build_panel(xframe, rframe)
        self.redraw(None)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def create_axes(self):
        gs = GridSpec(7,9)
        fig = self.fig
        ax10 = fig.add_subplot(gs[0:2,0])
        ax20 = fig.add_subplot(gs[2:,0])
        ax11 = fig.add_subplot(gs[0:2,1:6])
        axt1 = ax11.twinx()
        ax21 = fig.add_subplot(gs[2:,1:6])
        axt2 = ax21.twinx()
        axpn = fig.add_subplot(gs[1,6:9])

        self.axes = [ax10, ax20, ax11, ax21, axt1, axt2, axpn]

    def build_panel(self, xframe, rframe):

        # Column Parameters
        row = 0
        label = Tk.Label(xframe, text="Column Parameters", font=('MS Sans Serif', 16), bg='white')
        label.grid(row=row, column=0, columnspan=4)

        row += 1
        label = Tk.Label(xframe, text="Exclusion Limit", bg='white')
        label.grid(row=row, column=0, pady=5, sticky=Tk.E)

        self.excl_limit = Tk.DoubleVar()
        self.excl_limit.set(self.excl_limit_init)
        entry = Tk.Entry(xframe, textvariable=self.excl_limit, justify=Tk.CENTER, width=8)
        entry.grid(row=row, column=1, pady=5)

        label = Tk.Label(xframe, text="kDa", bg='white')
        label.grid(row=row, column=2, pady=5)

        button  = Tk.Button(xframe, text="MwRg Figure", command=self.show_mwrg_figure)
        button.grid(row=row, column=3, padx=20, pady=5)

        row += 1
        label = Tk.Label(xframe, text="Theoretical N of Plates", bg='white')
        label.grid(row=row, column=0, pady=5, sticky=Tk.E)

        self.num_plates_var = Tk.IntVar()
        self.num_plates_var.set(self.num_plates_disclosed)
        entry = Tk.Entry(xframe, textvariable=self.num_plates_var, justify=Tk.CENTER, width=8)
        entry.grid(row=row, column=1, pady=5)
        label = Tk.Label(xframe, text="N/m", bg='white')
        label.grid(row=row, column=2, pady=5)

        # SEC Parameters
        axpn = self.axes[6]
        axpn.set_axis_off()
        axpn.text(0.5, 0.2, "SEC Parameters", va="center", ha="center", fontsize=16)

        fig = self.fig

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(self.init_slider_specs, start=0):
            ax_ = fig.add_axes([0.73, 0.65 - 0.05*k, 0.2, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.label.set_size(16)
            # slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        self.slider_axes = slider_axes
        self.sliders = sliders

        # Rg and Proportions
        """
            import matplotlib
            print(matplotlib.rcParams['font.family'])
            which gave ['sans-serif']

            import tkinter
            print(tkinter.font.families())
        """

        for col, name in enumerate(["Comp. No", "Rg", "Proportion(%)", "τ/σ"]):
            label = Tk.Label(rframe, text=name, font=('MS Sans Serif', 16), bg='white')
            label.grid(row=0, column=col, padx=5)

        rg_vars = []
        pr_vars = []
        ts_vars = []
        for k, (rg, pr, ts) in enumerate(zip(self.init_rgs, self.init_prs, self.init_tss), start=1):
            label = Tk.Label(rframe, text="%d" % k, bg='white')
            label.grid(row=k, column=0, pady=5)

            var = Tk.DoubleVar()
            var.set(rg)
            rg_vars.append(var)
            sbox = SaferSpinbox(rframe, textvariable=var,
                            from_=10.0, to=80.0, format="%.1f", increment=0.1,
                            justify=Tk.CENTER, width=6)
            sbox.grid(row=k, column=1, pady=5)

            var = Tk.DoubleVar()
            var.set(pr)
            pr_vars.append(var)
            sbox = SaferSpinbox(rframe, textvariable=var,
                            from_=10.0, to=80.0, format="%.0f", increment=1.,
                            justify=Tk.CENTER, width=6)
            sbox.grid(row=k, column=2, pady=5)

            var = Tk.DoubleVar()
            var.set(ts)
            ts_vars.append(var)
            sbox = SaferSpinbox(rframe, textvariable=var,
                            from_=0.0, to=2.0, format="%.1f", increment=0.1,
                            justify=Tk.CENTER, width=6)
            sbox.grid(row=k, column=3, pady=5)

        self.rg_vars = rg_vars
        self.pr_vars = pr_vars
        self.ts_vars = ts_vars

        self.reset_axis = fig.add_axes([0.7, 0.05, 0.07, 0.07])
        self.reset_btn = Button(self.reset_axis, 'Reset')
        self.reset_btn.on_clicked(self.reset)

        self.redraw_axis = fig.add_axes([0.8, 0.05, 0.07, 0.07])
        self.refraw_btn = Button(self.redraw_axis, 'Redraw')
        self.refraw_btn.on_clicked(self.redraw)

        self.conform_axis = fig.add_axes([0.9, 0.05, 0.07, 0.07])
        self.conform_btn = Button(self.conform_axis, 'Conform')
        self.conform_btn.on_clicked(self.conform)

    def get_values_from_sliders(self):
        values = []
        for slider in self.sliders:
            values.append(slider.val)
        return values

    def show_mwrg_figure(self, debug=True):
        if debug:
            from importlib import reload
            import SecTheory.MwRgFigure
            reload(SecTheory.MwRgFigure)
        from SecTheory.MwRgFigure import MwRgFigure
        fig = MwRgFigure(self, self.excl_limit.get(), location="upper right")
        fig.show()

    def compute_draw_params(self, devel=True):
        if devel:
            from importlib import reload
            import SecTheory.SecParamsPlot
            reload(SecTheory.SecParamsPlot)
        from SecTheory.SecParamsPlot import compute_sec_peak_params

        rp, tI, t0, P, m = self.get_values_from_sliders()
        rgs = np.array([var.get() for var in self.rg_vars])
        Npc = self.num_plates_var.get() * 0.3   # 0.3m = 30cm

        trs, sigmas, xlim, x = compute_sec_peak_params(rp, tI, t0, P, m, rgs, Npc)

        props = np.array([var.get()/100 for var in self.pr_vars])
        ts_ratios = np.array([var.get() for var in self.ts_vars])

        # sigmas**2 == sig**2 + tau**2 = sig**2 * (1  + ts_ratio**2)
        #   where ts_ratio == tau/sig, tau = sig*ts_ratio
        # sig**2 = sigmas**2/(1  + ts_ratio**2)
        # sig = sigmas * sqrt(1/(1 + ts_ratio**2))
        # tau = sig*ts_ratio
        sig = sigmas * np.sqrt(1/(1 + ts_ratios**2))
        tau = ts_ratios * sig

        n_props = props/np.sum(props)

        hs = n_props
        model_params = np.vstack([hs, trs, sig, tau]).T

        def objective(p):
            model_params[:,0] = p
            areas = []
            for h, m, s, t in model_params:
                areas.append(np.sum(egh(x, h, m, s, t)))
            n_areas = areas/np.sum(areas)
            return np.sum((n_areas - n_props)**2)

        # get vertical scales which result in area proportions
        ret = minimize(objective, hs)
        model_params[:,0] = ret.x
        print("compute_draw_params", n_props, ret.x)

        return rgs, x, xlim, model_params, sigmas


    def reset(self, event):
        self.excl_limit.set(self.excl_limit_init)
        self.num_plates_var.set(self.num_plates_disclosed)

        for slider in self.sliders:
            slider.reset()

        for k, rg_var in enumerate(self.rg_vars):
            rg_var.set(round(self.init_rgs[k], 1))

        for k, pr_var in enumerate(self.pr_vars):
            pr_var.set(round(self.init_prs[k]))

        for k, ts_var in enumerate(self.ts_vars):
            ts_var.set(round(self.init_tss[k], 2))

        self.redraw(event)
        self.update()

    def redraw(self, event):
        rgs, x, xlim, model_params, sigmas = self.compute_draw_params()

        fig = self.fig

        fig.suptitle("Simple SEC Theory Simulation", fontsize=32)

        y_list = []
        for h, mu, s, tau in model_params:
            cy = egh(x, h, mu, s, tau)
            y_list.append(cy)
        ty = np.sum(y_list, axis=0)
        y_list.insert(0,ty)

        ax10, ax20 = self.axes[0:2]
        for ax, title in (ax10, "Entire\nView"), (ax20, "Peak\nView"):
            ax.set_axis_off()
            ax.text(0.5, 0.5, title, ha="center", va="center", fontsize=16)

        ax11, ax21 = self.axes[2:4]

        for ax in ax11, ax21:
            ax.cla()
            ax.grid(False)
            for k, y in enumerate(y_list):
                style = "-" if k == 0 else ":"
                ax.plot(x, y, style)

        ax21.set_xlim(*xlim)

        xmin, xmax = ax11.get_xlim()
        dx = (xmax - xmin)*0.05
        ymin, ymax = ax11.get_ylim()
        ax11.set_ylim(ymin, ymax)
        ty = ymin*0.7 + ymax*0.3
        label_fontsize = 14

        rp, tI, t0, P, m = self.get_values_from_sliders()
        tL = t0 + P
        k = 0
        for px, label in [(tI, "$t_I$"), (t0, "$t_0$"), (t0 + P, "$t_0 + P$")]:
            ax11.plot([px, px], [ymin, ymax], ":")
            px_ = (px - tI)/(tL - tI)
            dx_ = dx if k < 2 and px_ < 0.2 else -dx
            ax11.annotate(label, xy=(px, 0), xytext=(px + dx_, ty), ha='center',
                arrowprops=dict(arrowstyle="-", color='k'), fontsize=label_fontsize )
            k += 1

        rectangle = Rectangle(
                (xlim[0], ymin),    # (x,y)
                xlim[1] - xlim[0],  # width
                ymax - ymin,        # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax11.add_patch(rectangle)

        r = np.flip(np.arange(0, rp*0.9, 0.1))
        t = t0 + P*np.power(1 - r/rp, m)

        if event is not None:
            for k, ax in enumerate(self.axes[4:6]):
                ax.remove()
                self.axes[4+k] = self.axes[2+k].twinx()
        axts = self.axes[4:6]

        for k, axt in enumerate(axts):
            axt.set_ylabel("$R_g$", fontsize=label_fontsize)
            ms = 3 if k == 0 else 5
            axt.plot(t, r, "o", color="yellow", markersize=ms, label=r"$t_R = t_0 + P \cdot (1-\rho)^m$")
            axt.legend(fontsize=label_fontsize)

        axt = axts[1]
        axt.set_ylim(rgs[-1] - 2, rgs[0] + 2)

        xmin, xmax = axt.get_xlim()
        ymin, ymax = axt.get_ylim()
        dx = (xmax - xmin)*0.1
        dy = (ymax - ymin)*0.05

        trs = model_params[:,1]

        for rg, tr in zip(rgs, trs):
            rg_text = r"$R_g=%.1f$" % rg
            axt.annotate(rg_text, xy=(tr, rg), xytext=(tr + dx, rg + dy), ha='center',
                arrowprops=dict(arrowstyle="->", color='k'), fontsize=label_fontsize )

        if False:
            RgD = (rgs[0] - rgs[1])/(trs[1] - trs[0]) * np.average(sigmas)

            xmin, xmax = ax21.get_xlim()
            ymin, ymax = ax21.get_ylim()
            tx = xmin*0.8 + xmax*0.2
            ty = (ymin + ymax)/2
            ax21.text(tx, ty, "$RgD=%.3g$" % RgD, ha="center", va="center", fontsize=16)

        fig.tight_layout()
        # plt.show()

        if event is not None:
            self.mpl_canvas.draw()

    def get_lrf_info(self):
        if self.lrf_info is None:
            self.lrf_info = self.optimizer.objective_func(self.params, return_lrf_info=True)
        return self.lrf_info

    def get_gk_info(self, debug=True):
        if self.gk_info is None:
            if debug:
                from importlib import reload
                import Kratky.GuinierKratkyInfo
                reload(Kratky.GuinierKratkyInfo)
            from Kratky.GuinierKratkyInfo import GuinierKratkyInfo
            lrf_info = self.get_lrf_info()
            self.gk_info = GuinierKratkyInfo(self.optimizer, self.params, lrf_info)  # need_baseline=False
        return self.gk_info

    def conform(self, event, debug=True):
        if self.optimizer is None:
            raise Exception("Conform is not supported in Unit Test")

        gk_info = self.get_gk_info()
        print("rgs=", gk_info.rgs)
        for rg, var in zip(gk_info.rgs, self.rg_vars):
            var.set(round(rg,1))

        if debug:
            from importlib import reload
            import SecTheory.SecEstimator
            reload(SecTheory.SecEstimator)
        from SecTheory.SecEstimator import SecEstimator

        xr_curve = self.optimizer.xr_curve
        x = xr_curve.x
        y = xr_curve.y

        xr_params = self.separate_params[0]
        tI, Np_ignore = self.separate_params[8]
        Npm = self.num_plates_var.get()
        Npc = Npm*0.3

        seccol_params = self.separate_params[7]
        Npc_ignore, rp_ignore, tI, t0, P, m = seccol_params
        rp = self.sliders[0].val

        estimator = SecEstimator(Npc, rp, tI, t0, P, m)
        scaled_xr_params = self.lrf_info.get_scaled_xr_params(xr_params, debug=False)
        estimator.fit_to_decomposition(x, y, scaled_xr_params, gk_info.rgs)
        Npc, rp, tI, t0, P, m = estimator.get_params()
        print("ret: Npc, rp, tI, t0, P, m =",Npc, rp, tI, t0, P, m)
        for slider, val in zip(self.sliders, [rp, tI, t0, P, m]):
            slider.set_val(val)

        self.redraw(event)

def demo():
    from molass_legacy.KekLib.TkUtils import get_tk_root

    root = get_tk_root()
    root.withdraw()

    def demo_impl():
        dialog = SecSimulator(root)
        dialog.show()
        root.quit()

    root.after(0, demo_impl)
    root.mainloop()



import numpy                as np
import matplotlib.pyplot    as plt

y = 5 + 0.2 * np.random.randn( 10000 )

stdev = np.std( y )
print( 'std(y)=', stdev )

yn_stdev = np.std( [ np.average( y[i*10:(i+1)*10] ) for i in range(100) ] )

print( 'yn_stdev=', yn_stdev )
print( 'stdev/sqrt(10)=', stdev/np.sqrt(10)  )



"""
    TentativeDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

def demo(in_folder, sd):
    import molass_legacy.KekLib.DebugPlot as plt
    from importlib import reload
    import Models.ElutionCurveModels
    reload(Models.ElutionCurveModels)
    from .ElutionCurveModels import EGHA, EMGA

    print(in_folder)
    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.y

    if len(ecurve.peak_info) > 2:
        pno = 1
    else:
        pno = 0

    f, m, t = ecurve.peak_info[pno]
    x_ = x[f:t]
    y_ = y[f:t]

    model = EMGA()
    params = model.guess(y_, x=x_)
    out = model.fit(y_, params, x=x_)
    fy = model.eval(out.params, x=x)

    print(model.get_name())
    h, mu, sigma, tau, a = out.params
    print(sigma, tau)
    ratio = model.tau_bound_ratio
    print(ratio, sigma*ratio, tau)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title(in_folder)
        ax.plot(x, y)
        ax.plot(x_, y_)
        ax.plot(x, fy, ":")
        fig.tight_layout()
        plt.show()



"""
    Peaks.PeakEditorUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy.QuickAnalysis.ModeledPeaks import get_modeled_peaks_impl
from Selective.LrfSource import LrfSource

def apply_new_peaks(editor, new_peaks, debug=False):
    print("apply_new_peaks")

    a, b = editor.peak_params_set[2:4]
    uv_x, uv_y, xr_x, xr_y = editor.lrf_src_args1[0:4]
    num_peaks = len(new_peaks)

    # replace peak recognition with peaks=new_peaks
    uv_peaks, xr_peaks = get_modeled_peaks_impl(a, b, uv_x, uv_y, xr_x, xr_y, num_peaks, peaks=new_peaks, debug=debug)
    editor.peak_params_set.xr_peaks = xr_peaks
    editor.peak_params_set.uv_peaks = uv_peaks
    lrf_src = LrfSource(editor.sd, editor.corrected_sd, editor.lrf_src_args1, editor.peak_params_set)

    estimator = editor.fullopt.params_type.get_estimator(editor)
 
    init_params = estimator.estimate_params(lrf_src=lrf_src, debug=False)
    editor.draw_scores(init_params=init_params, create_new_optimizer=False)


# coding: utf-8
"""
    MinimumSurface.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from XrayData import XrayData
from MicrofluidicElution import get_mtd_elution
from SimpleUnfolding import proportion_folded
from SvdDenoise import get_denoised_data
import molass_legacy.KekLib.DebugPlot as plt

def compute_min_norm(G, m, x, M):
    pf = proportion_folded(G, m, x)
    pu = 1 - pf
    C = np.array([pf, x*pf, pu, x*pu])
    Cpinv = np.linalg.pinv(C)
    P = M@Cpinv
    z = np.linalg.norm(P@C - M)
    return z

def plot_surface(title, ax, x, GG, mm, M):
    zz = np.zeros(GG.shape)
    for i in range(GG.shape[0]):
        for j in range(GG.shape[1]):
            G_ = GG[i,j]
            m_ = mm[i,j]
            zz[i,j] = compute_min_norm(G_, m_, x, M)

    # demo_xy = [(1.74, 0.052), (7.89, 0.168), (54.2, 3.29)]
    # demo_xy = [(1.53e-09, 0.0519), (7.59e-10, 0.0524), (5.16, 0.107)]
    demo_xy = [     (0.,         0.05796411),
                    (0.99928439, 0.05584393),
                    (8.35120496, 0.09850686),
                    (8.72237253, 0.09875249),
                    (7.05725645, 0.10006241),
                    (5.95324888, 0.11418208),
                    (5.99807986, 0.1143187),
                    (6.99719508, 0.1000397),
                    (7.99302308, 0.18208686),
                    (9.00542046, 0.09865848),
                    ]
    demo_points = []
    for x_, y_ in demo_xy:
        z = compute_min_norm(x_, y_, x, M)
        demo_points.append((x_, y_, z))

    ax.set_title(title, fontsize=16)
    ax.set_xlabel('G')
    ax.set_ylabel('m')
    ax.set_zlabel('min norm')
    ax.plot_surface(GG, mm, zz, alpha=0.3)

    for x_, y_, z_ in demo_points:
        ax.plot([x_], [y_], [z_], 'o', markersize=10, label='(G,m)=(%.3g,%.3g)' % (x_,y_))

    ax.legend()

class MinimumSurface:
    def __init__(self, in_folder, gmax=12, mmax=0.3, auto=False, allq_only=True, save_no=None):
        xdata = XrayData(in_folder)
        num_files = xdata.data.shape[1]
        mtd_elution = get_mtd_elution(in_folder, num_files)
        slice_ = mtd_elution.guess_range(num_files)
        D = xdata.data[:,slice_]
        M = get_denoised_data(D, rank=4)

        """
        P = M@Cpinv
        """
        x = np.arange(M.shape[1])

        G = np.hstack([np.linspace(0.001, 0.09, 9), np.linspace(0.1, gmax, 50)])
        m = np.linspace(0.01, mmax, 50)

        GG, mm = np.meshgrid(G, m)
        print(GG.shape, mm.shape)

        if allq_only:
            fig = plt.figure(figsize=(15,9))
            ax2 = fig.add_subplot(111, projection='3d')
        else:
            fig = plt.figure(figsize=(18,8))
            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122, projection='3d')

        if not allq_only:
            i = xdata.e_index
            M_ = M[i,:].reshape((1,M.shape[1]))
            title = r'$ min \parallel P \cdot C_{G,m} - \tilde{M} \parallel $ Surface at Q=0.02'
            plot_surface(title, ax1, x, GG, mm, M_)
        title = r'$ min \parallel P \cdot C_{G,m} - \tilde{M} \parallel $ Surface for all Q'
        plot_surface(title, ax2, x, GG, mm, M)

        fig.tight_layout()
        plt.show(block=not auto)
        if auto:
            from time import sleep
            sleep(0.5)
        if save_no is not None:
            import os
            image_folder = r'.\temp\images'
            if not os.path.exists(image_folder) or save_no == 0:
                from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
                clear_dirs_with_retry([image_folder])
            path = image_folder + r'\image%03d.png' % save_no
            print([save_no], 'saving...')
            fig.savefig( path )

def make_it_into_a_movie(in_folder, G_range=(100, 12), m_range=(50, 0.3), num_frames=200):
    from subprocess import run
    gf, gt = G_range
    mf, mt = m_range

    gs = np.logspace(np.log10(gt), np.log10(gf), num_frames)
    ms = np.logspace(np.log10(mt), np.log10(mf), num_frames)

    k = 0
    for g, m in zip(reversed(gs), reversed(ms)):
        surface = MinimumSurface(in_folder, gmax=g, mmax=m, auto=True, allq_only=True, save_no=k)
        k += 1

    run(['ffmpeg', '-r', '8', '-i', r'temp\images\image%03d.png', '-vcodec', 'mpeg4', '-y', 'movie.mp4'])



# coding: utf-8
"""
    ElutionStainModel..py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""

class ElutionStainModel:
    def __init__( self, qvector, num_eluitions, start, stop ):
        assert num_eluitions > num_params

        self.qvec   = qvector
        self.j      = np.arange( num_eluitions )
        self.slice_ = slice( start, stop )
        self.num_params = stop - start

    def learn( self, baselines ):
        pass



"""
    ATSAS.AtsasUtils.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF
"""
import os
import glob
from .AutoRg import autorg_exe_array
from .AtsasVersion import get_atsas_bin_path    # for backward (or SAngler/CorMap) compatibility

def get_version(i=0):
    if len(autorg_exe_array) == 0:
        return None

    from molass_legacy.KekLib.OurImporter import import_module_from_path
    exe_path = autorg_exe_array[i]
    # print('exe_path=', exe_path)
    bin_dir, _ = os.path.split(exe_path)
    packages_pattern = bin_dir + r'\python*\site-packages'
    path_list = glob.glob(packages_pattern)
    if len(path_list) > 0:
        packages_path = path_list[0]
    else:
        atsas_dir, _ = os.path.split(bin_dir)
        packages_path = atsas_dir + r'\python2.7\site-packages'
        if not os.path.exists(packages_path):
             packages_path = None

    if packages_path is None:
        ver = 'Unkown'
    else:
        try:
            file_path = os.path.join(packages_path, "atsas.py")
            atsas = import_module_from_path('atsas', file_path)
            ver = atsas.ATSAS_VERSION
        except:
            print("loading from %s failed." % packages_path)
            ver = 'Failed'
    return ver

def get_versions():
    n = len(autorg_exe_array)
    if n == 0:
        return []
    vers = []
    for i in range(n):
        vers.append(get_version(i))
    return vers



"""
    PSD.PsdEstimator.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from molass_legacy.Models.Stochastic.TriporeColumn import TriporeColumn

def compute_psd(model, n, rgs, tRs):
    pore_sv = []
    pore_pv = []

def estimate_psd(model, n):
    pass


"""

    AnalysisRangeInfo.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

    This class is intended to unify the following existing range_info's and the newly
    introduced decomp_editor's range_info.

        from the attribute range_info
            AnalyzerDialog.range_info[0]
            AnalyzerDialog.set_analysis_ranges()
                AnalyzerUtil.set_analysis_ranges()

        from get_setting( ... )
            Analyzer.get_analysis_ranges()
                AnalyzerUtil.get_analysis_ranges()

    How to unify

        1)  define 'analysis_range_info' as a temporary setting info.
            this ensures that it is cleared everytime when the in_folder gets changed.

        2)  a range info is a list of range record, which has either one of
            the following styles

            [ PeakInfo, [f1, t1], [f2, t2] ]  (new style)
            [ PeakInfo, [f1, t1] ]            (new style)

            where old-style info has been convereted as follows

            [ lower, middle, upper ]  => [ PeakInfo(k, middle), [lower, middle], [middle, upper] ] 
            [ lower, upper, upper ]   => [ PeakInfo(k, upper), [lower, upper] ] for a microfluidic range

"""
import copy
import numpy as np
from molass_legacy.DataStructure.PeakInfo import PeakInfo
from molass_legacy._MOLASS.SerialSettings import set_setting

class PairedRange:
    def __init__(self, peak_info, range1=None, range2=None):
        """
        normally, range1 is not None.
        however, range1 can be None in abnormal cases as 20161104/BL-10C/OA
        """
        self.peak_id    = peak_info.peak_id
        self.top_x      = peak_info.top_x
        self.elm_recs   = peak_info.elm_recs
        self.fromto_list = []
        for r in [range1, range2]:
            if r is not None:
                self.fromto_list.append(r)

    def get_fromto_list(self):
        return self.fromto_list

    def as_list(self, k):
        return [PeakInfo(k, self.top_x, self.elm_recs)] + self.fromto_list

    def __repr__(self):
        return 'PairedRange(%s,%s)' % ("PeakInfo(%d,%g,'elm_recs')" % (self.peak_id,self.top_x),
                                       ",".join([str(rec) for rec in self.fromto_list]))

    def __getitem__(self, index):
        return self.fromto_list[index]

    def get_log_str(self):
        if len(self.fromto_list) == 1:
            return 'range ' + str(self.fromto_list[0])
        else:
            return 'range pair ' + str(self.fromto_list)

    def update_range(self, ad, f, t):
        r = self.fromto_list[ad]
        r[0] = f
        r[1] = t

    def get_concatenated_range(self):
        return self.fromto_list[0][0], self.fromto_list[-1][-1]

def upgrade_ranges(ranges):
    ret_ranges =[]
    for k, range_ in enumerate(ranges):
        if range_.__class__.__name__ == "PairedRange":  # type(range_) seems to require a qualifier, i.e., AnalysisRangeInfo.PairedRange
            range_ = range_.as_list(k)
        if len(range_) < 2:
            # currently, this case happens only from make_default_analysis_range_info
            # with select_fix
            continue
        if np.isscalar(range_[1]):
            lower, middle, upper = range_
            info_ = PeakInfo( k, middle )
            if middle == upper:
                # i.e., for a microfluidic range
                range_ = [ info_, [lower, upper] ]
            else:
                range_ = [ info_, [lower, middle], [middle, upper] ]
            ret_ranges.append(range_)
        else:
            ret_ranges.append(range_)
    return ret_ranges

class AnalysisRangeInfo:
    def __init__(self, ranges, editor=None):
        self.ranges = upgrade_ranges(ranges)
        self.editor = editor

    def get_ranges(self):
        return self.ranges

    def get_old_style_ranges(self):
        ranges = []
        for range_ in self.ranges:
            n = len(range_)
            if n == 3:
                left_rec = range_[1]
                right_rec = range_[2]
                ranges.append( left_rec + [right_rec[1]] )
            elif n == 2:
                lower, upper = range_[1]
                middle = (lower + upper)//2
                ranges.append( [lower, middle, upper] )
            else:
                assert False
        return ranges

    def __repr__(self):
        return str(self.ranges)

def report_ranges_from_analysis_ranges(j0, analysis_ranges):
    return [ [ [j0+f, j0+t] for f, t in prange.get_fromto_list()] for prange in analysis_ranges ]

def convert_to_paired_ranges(in_ranges):
    # print( 'convert_to_paired_ranges: in_ranges=', in_ranges )

    ranges = upgrade_ranges(in_ranges)

    paired_ranges = []
    num_ranges = 0

    for k, range_ in enumerate(ranges):
        if len(range_) == 2:
            prange_ = PairedRange(range_[0], range_[1])
        else:
            prange_ = PairedRange(range_[0], range_[1], range_[2])
        paired_ranges.append( prange_ )
        num_ranges += len( prange_.fromto_list )

    return paired_ranges, num_ranges

def shift_paired_ranges(j0, paired_ranges):
    ret_ranges = []
    for r in paired_ranges:
        new_r = copy.deepcopy(r)
        new_r.fromto_list = [ [ ft-j0  for ft in rec] for rec in r.fromto_list]
        ret_ranges.append(new_r)
    return ret_ranges

def shift_range_from_to_by_x(x, f, t):
    return [min(max(0, int(v - x[0])), len(x)-1) for v in [f, t]]

def get_analysis_info_from_curve(curve, editor):
    paired_ranges = curve.get_default_paired_ranges()
    return AnalysisRangeInfo(paired_ranges, editor)

def set_default_analysis_range_info(x_curve):
    from RangeInfo import RangeEditorInfo
    paired_ranges = x_curve.get_default_paired_ranges()
    set_setting( 'range_type', 4 )
    analysis_range_info = AnalysisRangeInfo(paired_ranges, editor='RangeEditor')
    set_setting( 'analysis_range_info', analysis_range_info )
    ignorable_flags = [0] * len(paired_ranges)
    set_setting( 'range_editor_info', RangeEditorInfo(paired_ranges, ignorable_flags))



"""
    FunctionChanger.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import re
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from .FuncImporter import get_objective_function_info

func_code_re = re.compile(r"(F\d{4})")

class FunctionChanger(Dialog):
    def __init__(self, parent, js_canvas):
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.js_canvas = js_canvas
        Dialog.__init__(self, parent, "Function Changer", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        class_code = self.js_canvas.dialog.class_code

        func_info = get_objective_function_info(logger=self.logger, default_func_code=class_code)
        self.func_dict = func_info.func_dict
        self.key_list = []
        for key in func_info.key_list:
            m = func_code_re.search(key)
            if m:
                func_code = m.group(1)
                if func_code[0:3] == class_code[0:3]:
                    self.key_list.append(key)

        self.selected_function = Tk.StringVar()
        self.selected_function.set(self.key_list[func_info.default_index])
        self.function_box = ttk.Combobox(master=body_frame, values=self.key_list, textvariable=self.selected_function,
                                width=80, justify=Tk.CENTER, state=Tk.NORMAL)
        self.function_box.pack(side=Tk.LEFT)

    def apply(self):
        func_key = self.selected_function.get()
        m = func_code_re.search(func_key)
        class_code = m.group(1)

        js_canvas = self.js_canvas
        js_canvas.dialog.class_code = class_code
        js_canvas.draw_suptitle()
        js_canvas.mpl_canvas.draw()

        func_class = self.func_dict[func_key]
        state_info = js_canvas.dialog.state_info
        state_info.optinit_info.class_code = class_code
        state_info.optinit_info.fullopt_class = func_class



# coding: utf-8
"""

    AnimObjects.py

        recognition of peaks

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
import copy

class AnimMapper:
    def __init__( self, mapper ):
        self.mapper     = mapper
        self.opt_params = mapper.opt_params
        self.a_curve    = mapper.a_curve
        self.a_spline   = mapper.a_spline
        self.a_vector   = mapper.a_vector
        self.a_base     = mapper.a_base
        self.a_base_adjustment   = 0
        self.flow_changes   = mapper.flow_changes
        self.x_curve    = mapper.x_curve
        self.x_spline   = mapper.x_spline
        self.x_vector   = mapper.x_vector
        self.x_base     = mapper.x_base
        self.x_base_adjustment   = 0
        self.x_ranges   = None
        self.uv_ranges  = None
        self.mapping_ranges = None
        self.opt_results    = None
        self.mapped_vector  = None
        self.mapped_spline  = None
        self.a_base_adjustment  = 0
        self.x_base_adjustment  = 0
        self.inv_mapped_boundaries  = None
        self.x_curve_y_adjusted = None
        self.in_opt = False
        self.in_adj = False
        self.opt_phase = 0

    def copy( self, deep_ranges=False, with_adj=False ):
        mapper  = self.mapper
        obj = AnimMapper( mapper )
        obj.opt_phase = self.opt_phase      # opt_phase of the copy source
        try:
            if deep_ranges:
                obj.x_ranges        = copy.deepcopy( mapper.x_ranges )
                obj.uv_ranges       = copy.deepcopy( mapper.uv_ranges )
                obj.mapping_ranges  = copy.deepcopy( mapper.mapping_ranges )
            else:
                obj.x_ranges        = mapper.x_ranges
                obj.uv_ranges       = mapper.uv_ranges
                obj.mapping_ranges  = mapper.mapping_ranges
        except:
            pass
        try:
            if with_adj:
                obj.a_base_adjustment   = copy.deepcopy( mapper.a_base_adjustment )
                obj.x_base_adjustment   = copy.deepcopy( mapper.x_base_adjustment )
            obj.opt_results     = mapper.opt_results
        except:
            pass

        try:
            obj.mapped_vector   = mapper.mapped_vector
            obj.mapped_spline   = mapper.mapped_spline
        except:
            pass

        try:
            obj.x_curve_y_adjusted      = mapper.x_curve_y_adjusted
        except:
            pass

        try:
            obj.inv_mapped_boundaries   = mapper.inv_mapped_boundaries
        except:
            pass

        return obj



"""
    Simulative/PsdSimulation.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
from importlib import reload
import Batch.DataBridgeUtils
reload(Batch.DataBridgeUtils)
from molass_legacy.Batch.DataBridgeUtils import get_lrf_source_impl

def pds_simulation_impl(editor, show_fixed_demo=False):
    print("psd_simulation_impl")
    if show_fixed_demo:
        from importlib import reload
        import Simulative.SimulativeDemo
        reload(Simulative.SimulativeDemo)
        from Simulative.SimulativeDemo import demo
        # showing temporarily the fixed demo
        demo(parent=editor.parent)
    else:
        parent = editor
        parent.config(cursor="wait")
        parent.update()
        lrf_src = get_lrf_source_impl(editor)
        lrf_src.run_simulation_process(editor.parent)
        parent.config(cursor="")


"""
    Processes.SharedInfo.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from multiprocessing import Value, Array, RawArray

NUM_QUEUES = 4
QUEUE_GUI = 0
QUEUE_LOADER = 1
QUEUE_RG = 2
QUEUE_RECOG = 3
BUFFER_SIZE = 500*1500*3

class SharedString:
    def __init__(self, s="", maxlen=256):
        self.maxlen = Value('i', maxlen)
        self.strlen = Value('i', 0)
        self.buffer = Array('u', maxlen)
        self.set(s)

    def set(self, s):
        assert len(s) <= self.maxlen.value
        self.strlen.value = len(s)
        self.buffer[:self.strlen.value] = s

    def __str__(self):
        return ''.join(self.buffer[:self.strlen.value])

class SharedInfo:
    def __init__(self):
        self.value = Value('i', 0)
        self.in_folder = SharedString()
        self.proc_states = Array('i', np.zeros(10, dtype=int))
        self.buffer = RawArray('d', BUFFER_SIZE)
        self.buffer_shape = Array('i', [0,0,0])
        self.X = None
        self.curve_buffer = None
        self.curve_xy = None
        self.max_y = Value('d', 0)

    def __repr__(self):
        return str((self.value.value, str(self.in_folder), [v for v in self.proc_states]))

    def set_procstate(self, pn, state):
        self.proc_states[pn] = state

    def get_procstate(self, pn):
        return self.proc_states[pn]

    def initilize_buffer(self, shape=None, array=None, ecurve=None):
        if shape is None:
            assert array is not None
            shape = array.shape
        for i, n in enumerate(shape):
            self.buffer_shape[i] = n
        size = np.prod(shape)
        assert size <= BUFFER_SIZE
        self.X = np.frombuffer(self.buffer, count=size).reshape(shape)
        if array is None:
            array = np.zeros(shape)
        np.copyto(self.X, array)
        print("initilize_buffer:", shape, size)
        if ecurve is not None:
            size = len(ecurve.x)*2
            self.curve_buffer = RawArray('d', size)
            self.curve_xy = np.frombuffer(self.curve_buffer, count=size).reshape((2,size//2))
            self.max_y.value = ecurve.max_y
            np.copyto(self.curve_xy, np.array([ecurve.x, ecurve.y]))

    def set_xr_data(self, i, data):
        self.X[i,:,:] = data

    def get_buffer_ready(self):
        shape = [n for n in self.buffer_shape]
        size = np.prod(shape)
        assert size <= BUFFER_SIZE
        self.X = np.frombuffer(self.buffer, count=size).reshape(shape)
        print("get_buffer_ready:", shape, size)
        if self.curve_buffer is not None:
            size = self.X.shape[0]
            self.curve_xy = np.frombuffer(self.curve_buffer, count=size*2).reshape((2, size))

    def get_xr_data(self, i):
        return self.X[i,:,:]



# coding: utf-8
"""

    HyperbolaDemo.py

    Copyright (c) 2021, SAXS Team, KEK-PF

"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from .Hyperbola import RotatedHyperbola

def demo():
    from .ElutionModels import gaussian

    x = np.linspace(0, 200, 201)
    height = 3
    gy = gaussian(x, height, 100, 20)
    peak_slice = slice(80, 120)

    fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21,11))
    fig.suptitle("Rotated Hyperbola Local Peak Model", fontsize=20)

    for i, angles in enumerate([[0], [0, -3, 3]]):
        for ax, a in zip(axes[i,:], [0.2, 1, 5]):
            ax.set_title("a=%g (related to curvature)" % a, fontsize=16)
            ax.plot(x, gy, lw=3, label="gaussian")
            for r in angles:
                ptx, pty= 100,height
                rh = RotatedHyperbola(ptx, pty, a=a, hbw=30, deg=r)
                # rh.fit(x[peak_slice], gy[peak_slice])
                ax.plot(x, rh(x), label="hyperbola rotation(%g°)" % r)
                ax.plot(ptx, pty, 'o', color='red')
                ax.set_ylim(-2, 4)
            ax.legend(fontsize=12)
    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()

def demo_real(in_folder, pno=0):
    from matplotlib.patches import Rectangle
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Peaks.RobustPeaks import RobustPeaks
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd(proxy=False)
    xr_curve = sd.get_xray_curve()
    x = xr_curve.x
    y = xr_curve.y
    rp = RobustPeaks(x, y)
    ls, pt, rs = rp.get_peaks()[pno]
    print((ls, pt, rs))
    hw = (pt-ls + rs-pt)//4
    by = np.min(y[[pt-hw, pt+hw]])
    ls_ = bisect_right(y[0:pt], by)
    rs_ = pt + bisect_right(-y[pt:], -by)
    slice_ = slice(pt-hw, pt+hw)
    # slice_ = slice(ls_, rs_)
    x_ = x[slice_]
    y_ = y[slice_]

    if True:
        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x_, y_)
        ax.plot(x[pt],y[pt], 'o', color='red')
        plt.show()
        plt.pop()

    # fx, fy = x[ls:rs], y[ls:rs]
    fx, fy = x_, y_

    try:
        hbw = ((x[pt] - x[ls]) + (x[rs] - y[pt]))
        rh = RotatedHyperbola(x[pt], y[pt], hbw=hbw)
        rh.fit(fx, fy)
        success = True
    except:
        success = False
    fig, ax = plt.subplots()
    ax.plot(x, y)
    if False:
        for rad in [1/2]:
            ax.plot(x_, rot_hyperbola(x_, 1.1, 1, 0, 100, 1.5, rad*np.pi))

    if success:
        ax.plot(fx, rh(fx))
        ptx, pty = rh.get_peak_top()
        ax.plot(ptx, pty, 'o', color='red')

    ymin, ymax = ax.get_ylim()
    ymax_ = ymax*1.1
    ax.set_ylim(ymin, ymax_)

    f, t = x[[ls,rs]]
    p = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax_ - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )
    ax.add_patch(p)

    plt.show()



"""
    Theory.Survey.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""

from molass_legacy.Batch.StandardProcedure import StandardProcedure
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from .SphericalFit import compute_bq_score

def compute_bq_scores_for_all_peaks(sd):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    ret_list = []
    for pno, range_ in enumerate(ecurve.get_ranges_by_ratio(0.5)):
        f, p, t = range_

        eslice = slice(f,t)

        if pno == ecurve.primary_peak_no:
            gf, gt = sd.pre_recog.get_gunier_interval()
            preRg = sd.pre_recog.get_rg()
        else:
            pre_rg = sd.pre_recog.pre_rg
            preRg = pre_rg.compute_rg(selected=p)
            sg = pre_rg.sg
            gf = sg.guinier_start
            gt = sg.guinier_stop
        gslice = slice(gf, gt)

        ret = compute_bq_score(M, E, qv, ecurve, gslice, eslice, preRg)
        print([pno], p, ret[-1])
        ret_list.append(ret)

    return ret_list

def get_sd(in_folder):
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd(debug=False)
    pre_recog = PreliminaryRecognition(sd)
    sd_ = pre_recog.get_analysis_copy()
    return sd_
    compute_bq_scores_for_all_peaks(sd_)

def bq_scores_survey_for_a_folder(in_folder, fh=None):
    from molass_legacy._MOLASS.SerialSettings import clear_temporary_settings
    from DataUtils import cut_upper_folders

    print('Doing', in_folder)
    clear_temporary_settings()      # iterating without this can be buggy! be careful.

    try:
        sd = get_sd(in_folder)
        for pno, ret in enumerate(compute_bq_scores_for_all_peaks(sd)):
            if fh is None:
                print([pno], ret)
            else:
                fh.write(','.join([cut_upper_folders(in_folder), str(pno)] + ['%g' % v for v in ret]) + '\n')
                fh.flush()
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
        etb = ExceptionTracebacker()
        print(etb.last_lines())

def bq_scores_survey():
    from DataUtils import get_pytools_folder, serial_folder_walk

    fh = open('bq_scores.csv', 'w')

    def proc(in_folder, uv_folder, plot=False):
        bq_scores_survey_for_a_folder(in_folder, fh=fh)
        return True, None

    data_folder = get_pytools_folder() + '/Data'
    serial_folder_walk(data_folder, proc)
    fh.close()



"""
    SecTheory.PoresizeFreedom.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from DataUtils import get_in_folder
from ModelParams.EghParams import EghAdvansedParams
from .RetensionTime import compute_retention_time

def freedom_plot(sseq):
    in_folder = get_in_folder(sseq.get_in_folder().replace("\\", "/"))

    fv_array, x_array = sseq.get_info()[0:2]
    k = np.argmin(fv_array[:,1])
    pasrams_type = EghAdvansedParams(sseq.get_nc())
    separated = pasrams_type.split_params_simple(x_array[k,:])
    xr_params = separated[0]
    rg_params = separated[2][:-1]
    seccol_params = separated[-1]

    f_trs = xr_params[:,1]

    fig, axes = plt.subplots(ncols=5, figsize=(20, 4))
    fig.suptitle("Demo indicating Freedom of Pore Size using %s" % in_folder, fontsize=20)

    t0, k, rp, m = seccol_params

    poresizes = [60, 70, 80, 90, 100]
    fit_scores = []
    params_list = []
    i = 0
    for ax, ps in zip(axes, poresizes):
        ax.set_title("Pore Size=%.3g" % ps, fontsize=16)
        ax.plot(f_trs, rg_params, "-o", label="current fit")

        secc_params, min_ret = adjust_to_poresize(f_trs, rg_params, seccol_params, ps)
        params_list.append(secc_params)
        fit_scores.append(min_ret.fun)

        m_trs = compute_retention_time(secc_params, rg_params)
        ax.plot(m_trs, rg_params, "-o", label="stochastic model")
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        tx = (xmin + xmax)/2
        ty = (ymin + ymax)/2
        ax.text(tx, ty, ", ".join(["%.3g" % v for v in secc_params]), ha="center", alpha=0.5, fontsize=16)
        ax.legend()
        i += 1

    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    plt.show()

    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,6))
    fig.suptitle("SEC Parameters Dependence on Pore Size using %s" % in_folder, fontsize=20)

    ax1.set_title("Comparison of Curves with Different Poresizes", fontsize=16)
    ax1.set_xlabel("Retention Time", fontsize=16)
    ax1.set_ylabel("Rg", fontsize=16)

    ax1.plot(f_trs, rg_params, "-o", label="current fit")
    wider_rgs = np.linspace(10, 120, 100)
    for ps, params in zip(poresizes, params_list):
        m_trs = compute_retention_time(params, wider_rgs)
        ax1.plot(m_trs, wider_rgs, ":", label="poresize %g" % ps)
    ax1.legend()

    params_array = np.array(params_list)

    axt = ax2.twinx()
    axt.grid(False)

    ax2.set_title("t0, T, rp, m Variation vs. Poresizes", fontsize=16)
    ax2.set_xlabel("Pore Size", fontsize=16)

    ax2.plot(poresizes, params_array[:,0], label="t0")
    ax2.plot(poresizes, params_array[:,1], label="T")
    ax2.plot(poresizes, params_array[:,2], label="rp")
    axt.plot(poresizes, params_array[:,3], ":", label="m")
    ax2.legend()
    axt.legend()

    ax3.set_title("Fit Scores vs. Poresize", fontsize=16)
    ax3.set_xlabel("Pore Size", fontsize=16)
    ax3.set_ylabel("Fit Score", fontsize=16)

    ax3.plot(poresizes, fit_scores, "o")

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()

def adjust_to_poresize(f_trs, rgs, init_params, ps):
    x_ = np.zeros(4)
    x_[2] = ps
    def objective_func(x):
        x_[[0,1,3]] = x
        model_trs = compute_retention_time(x_, rgs)
        return np.sqrt(np.average((model_trs - f_trs)**2))

    ret = minimize(objective_func, init_params[[0,1,3]])
    x_[[0,1,3]] = ret.x

    return x_, ret



# coding: utf-8
"""
    SmoothnessPenalty.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""

import numpy as np

try:
    from SmoothnessPenaltyNumba import *
    NUMBA_IS_AVAILABLE  = True
except:
    NUMBA_IS_AVAILABLE  = False

assert NUMBA_IS_AVAILABLE

import molass_legacy.KekLib.DebugPlot as plt

"""
    [NumPy k-th diagonal indices]
        (https://stackoverflow.com/questions/10925671/numpy-k-th-diagonal-indices)
"""
def kth_diag_indices(a, k):
    rowidx, colidx = np.diag_indices_from(a)
    colidx = colidx.copy()  # rowidx and colidx share the same buffer

    if k > 0:
        colidx += k
    else:
        rowidx -= k
    k = np.abs(k)

    return rowidx[:-k], colidx[:-k]


"""
    http://ttic.uchicago.edu/~klivescu/papers/wang_SLT2014.pdf
    RECONSTRUCTION OF ARTICULATORY MEASUREMENTS
    WITH SMOOTHED LOW-RANK MATRIX COMPLETION
"""
def norm_diff_matrix(n):
    dm = np.identity(n)*2
    dm[0,0] = 1
    dm[-1,-1] = 1
    dm[kth_diag_indices(dm,1)] = -1
    dm[kth_diag_indices(dm,-1)] = -1
    return dm

def diff_matrix(n):
    dm = np.identity(n)
    dm[kth_diag_indices(dm,1)] = -1
    return dm

def get_penalty_impl(NDM, P, penalty_weights):
    return [P[:,j].T @ NDM @ P[:,j] * penalty_weights[j%2]   for j in range(P.shape[1]) ]

def get_penalty_diff_impl(NDM, P, penalty_weights):
    n = P.shape[1]
    NDP = np.array( [ ( (NDM * 2) @ P[:,j] * penalty_weights[j%2] ).T  for j in range(n) ] ).T
    return NDP

class SmoothnessPenalty:
    def __init__(self, m, conc_dependence=2, add_const=0):
        self.NDM = norm_diff_matrix(m)
        self.conc_dependence = conc_dependence
        self.add_const = add_const

    def get_reshaped(self, flatP):
        m = self.NDM.shape[0]
        assert len(flatP) % m == 0
        n = len(flatP)//m
        self.col_size = n - self.add_const
        return flatP.reshape((m, n))

    def get_penalties(self, flatP, penalty_weights):
        P = self.get_reshaped(flatP)
        if NUMBA_IS_AVAILABLE:
            return get_penalty_impl_numba(self.conc_dependence, P, penalty_weights, self.col_size)
        else:
            return get_penalty_impl(self.NDM, P, penalty_weights)

    def get_penalties_bq_ignore(self, flatP, penalty_weight):
        P = self.get_reshaped(flatP)
        return get_penalty_bq_ignore_impl_numba(P, penalty_weight, self.col_size)

    def get_penalty_diff(self, flatP, penalty_weights, debug=False):
        P = self.get_reshaped(flatP)
        m = self.NDM.shape[0]
        n = P.shape[1]
        if NUMBA_IS_AVAILABLE:
            NDP = get_penalty_diff_impl_numba(self.conc_dependence, P, penalty_weights, self.col_size)
        else:
            NDP = get_penalty_diff_impl(self.NDM, P, penalty_weights)

        if debug:
            DM = diff_matrix(m)
            DP = np.array( [ ( DM @ P[:,j] ).T  for j in range(n) ] ).T

            plt.plot(P)
            # plt.plot(DP)
            plt.plot(NDP)
            plt.show()

        return NDP.flatten()

    def get_penalty_diff_bq_ignore(self, flatP, penalty_weights, debug=False):
        P = self.get_reshaped(flatP)
        NDP = get_penalty_diff_bq_ignore_impl_numba(P, penalty_weights, self.col_size)
        return NDP.flatten()



# coding: utf-8
"""
    CallStack.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import re
import os
import traceback

class CallStack:
    def __init__(self, terse=True):
        stack = traceback.format_stack()[:-2]
        if terse:
            c_seq = []
            file_re = re.compile(r'File "(.+)"')
            for line in stack:
                c_seq.append( re.sub(file_re, lambda m: 'File "' + os.path.split(m.group(1))[1] + '"', line) )
            self.c_seq = c_seq
        else:
            self.c_seq = stack

    def __str__(self):
        return ''.join(self.c_seq)

    def __repr__(self):
        return self.__str__()



"""
    SimData.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
from bisect import bisect_right
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from scipy.interpolate import UnivariateSpline, LSQUnivariateSpline
from molass_legacy.DENSS.DenssUtils import fit_data_impl
from molass_legacy.Models.ElutionCurveModels import emg

class SimData:
    def __init__(self, a_file, b_file, num_elutions=300,  debug=False):
        A = np.loadtxt(a_file)
        q = A[:,0]
        a = A[:,1]
        e = A[:,2]
        sasrec, work_info = fit_data_impl(q, a, e, file=a_file, D=76)
        qc = sasrec.qc
        ac = sasrec.Ic
        ec = work_info.Icerr
        spline = UnivariateSpline(qc, ac, s=0)
        self.q = q
        self.a = spline(q)
        self.ae = e

        B = np.loadtxt(b_file)
        q = B[:,0]
        b = B[:,1]
        e = B[:,2]
        num_nkots = 10
        knots = np.linspace( 0, q[-1], num_nkots + 2 )
        spline = LSQUnivariateSpline(q, b, knots[1:-1], ext=3)
        self.b = spline(q)
        self.be = e

        self.i = bisect_right(q, 0.02)

        num_elutions = 300
        h = 1
        mu = num_elutions//2
        sigma = 30
        tau = 0
        self.cx = x = np.arange(num_elutions)
        self.e_params = [(h, mu, sigma, tau)]
        self.cy = emg(x, h, mu, sigma, tau)

        if debug:
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))
            ax1.set_yscale('log')
            ax1.plot(q, a)
            ax1.plot(qc, ac)

            ax2.set_yscale('log')
            ax2.plot(q, a)
            ax2.plot(q, self.a)

            ax3.plot(q, b)
            ax3.plot(q, self.b)

            fig.tight_layout()
            plt.show()

    def get_data(self, noise=0.01, conc=0, seed=None):
        c = self.cy/np.max(self.cy)
        if conc > 0:
            P = np.array([self.a, self.b*conc]).T
            C = np.array([c, c**2])
        else:
            P = self.a[:, np.newaxis]
            C = c[np.newaxis,:]

        M = P @ C
        print('M.shape', M.shape)
        M_ = M/np.max(M)

        if seed is not None:
            np.random.seed(seed)

        if noise > 0:
            GE = np.random.normal(0, 1, M.shape)*noise/M_
            M *= (1 + GE)
            E = np.abs(M*GE)

        return M, E

    def demo_plot(self, axes=None, show=True):
        from MatrixData import simple_plot_3d
        from SvdDenoise import get_denoised_data
        from molass_legacy.Models.ElutionCurveModels import emg_x_from_height_ratio
        from matplotlib.patches import Rectangle

        noise = 0.01
        cd = 1
        M, E = self.get_data(noise=noise, conc=cd)
        c = self.cy/np.max(self.cy)
        C = np.array([c, c**2])

        h, mu, sigma, tau = self.e_params[0]
        f, t = [int(p+0.5) for p in emg_x_from_height_ratio(0.5, mu, sigma, tau)]
        print("(f, t)=", (f, t))
        eslice = slice(f,t+1)
        M_ = get_denoised_data(M[:,eslice], rank=2)
        C_ = C[:,eslice]
        P_ = M_ @ np.linalg.pinv(C_)

        if axes is None:
            fig = plt.figure(figsize=(16, 11))
            ax0 = fig.add_subplot(221)
            ax1 = fig.add_subplot(222)
            ax2 = fig.add_subplot(223, projection='3d')
            ax3 = fig.add_subplot(224)

        else:
            ax1, ax2, ax3 = axes

        ax0.set_title("Gaussian Elution Curve", fontsize=16)
        ax0.plot(self.cx, self.cy, color='orange')

        ax1.set_title("Smoothed Extrapolated Curves from 20180526/OA", fontsize=16)
        ax1.set_yscale('log')
        ax1.plot(self.q, self.a, color='C1', label='A(q)')

        ax1t = ax1.twinx()
        ax1t.grid(False)
        ax1t.plot(self.q, self.b, color='pink', label='B(q)')

        ax1.legend()
        ax1t.legend(bbox_to_anchor=(1, 0.95), loc='upper right')

        ax2.set_title("Generated Data with Noise=%g, SCD=%d" % (noise, cd), fontsize=16)
        simple_plot_3d(ax2, M)

        ymin, ymax = ax0.get_ylim()
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax0.add_patch(p)

        ax3.set_title("Solved Extrapolated Curves", fontsize=16)
        ax3.set_yscale('log')
        ax3.plot(self.q, P_[:,0], color='C1', label='A(q)')

        ax3t = ax3.twinx()
        ax3t.grid(False)
        ax3t.plot(self.q, P_[:,1], color='pink', label='B(q)')

        ax3.legend()
        ax3t.legend(bbox_to_anchor=(1, 0.95), loc='upper right')

        fig.tight_layout()

        if show:
            plt.show()
